# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module contains unit operations for the separation of solids 
(e.g. centrifugation, expression, filtration).

.. contents:: :local:
    
.. autoclass:: biosteam.units.solids_separation.SolidsSeparator
.. autoclass:: biosteam.units.solids_separation.SolidsCentrifuge
.. autoclass:: biosteam.units.solids_separation.RotaryVacuumFilter 
.. autoclass:: biosteam.units.solids_separation.PressureFilter
.. autoclass:: biosteam.units.solids_separation.ScrewPress

References
----------
.. [1] Seider, Warren D., et al. (2017). "Cost Accounting and Capital Cost 
    Estimation". In Product and Process Design Principles: Synthesis, Analysis,
    and Evaluation (pp. 481-485). New York: Wiley.
.. [2] Humbird, D., Davis, R., Tao, L., Kinchin, C., Hsu, D., Aden, A.,
    Dudgeon, D. (2011). Process Design and Economics for Biochemical 
    Conversion of Lignocellulosic Biomass to Ethanol: Dilute-Acid 
    Pretreatment and Enzymatic Hydrolysis of Corn Stover
    (No. NREL/TP-5100-47764, 1013269). https://doi.org/10.2172/1013269

"""
from .decorators import cost
from .splitting import Splitter
from .design_tools import compute_vacuum_system_power_and_cost
from warnings import warn
from ..exceptions import lb_warning, InfeasibleRegion
from .decorators import cost
from biosteam.utils import remove_undefined_chemicals, default_chemical_dict
import numpy as np
import biosteam as bst
from thermosteam import separations
from math import exp, log, ceil

__all__ = ('SolidsSeparator', 'RotaryVacuumFilter', 'CrushingMill', 
           'PressureFilter', 'SolidsCentrifuge', 'RVF',
           'ScrewPress',)

class SolidsSeparator(Splitter):
    """
    Create SolidsSeparator object.
    
    Parameters
    ----------
    ins : 
        Inlet fluids with solids.
    outs : 
        * [0] Retentate.
        * [1] Permeate.
    split : array_like
        Component splits to 0th output stream
    moisture_content : float
        Fraction of water in solids
    
    """
    _N_ins = 1
    _ins_size_is_fixed = False
    
    def _init(self, split, 
            order=None, moisture_content=None, 
            moisture_ID=None,
            strict_moisture_content=None
        ):
        Splitter._init(self, order=order, split=split)
        #: Moisture content of retentate
        self.moisture_content = moisture_content
        self.strict_moisture_content = strict_moisture_content
        if moisture_content is not None:
            if moisture_ID is None: moisture_ID = '7732-18-5'
            self.moisture_ID = moisture_ID
    
    def _run(self):
        if self.moisture_content is None:
            separations.mix_and_split(
                self.ins, *self.outs, self.split,
            )
        else:
            moisture_ID = self.moisture_ID
            self.isplit[moisture_ID] = 0.
            separations.mix_and_split_with_moisture_content(
                self.ins, *self.outs, self.split, self.moisture_content, self.moisture_ID,
                self.strict_moisture_content,
            )
    #     if self._recycle_system and self._system.algorithm == 'Phenomena oriented':
    #         ID = self.moisture_ID
    #         if not ID: return
    #         top, bottom = self.outs
    #         top_mol = top.imol[ID]
    #         self.isplit[ID] = top_mol / (top_mol + bottom.imol[ID])
            
    # def _update_nonlinearities(self):
    #     outs = self.outs
    #     data = [i.get_data() for i in outs]
    #     self._run()
    #     for i, j in zip(outs, data): i.set_data(j)


class SolidsCentrifuge(SolidsSeparator):
    """
    Create a solids centrifuge that separates out solids according to
    user defined split. Capital cost is based on [1]_.
    
    Parameters
    ----------
    ins : 
        Inlet fluid with solids.
    outs : 
        * [0] Solids-rich stream.
        * [1] Liquid-rich stream.
    split : array_like or dict[str, float]
           Component splits.
    order=None : Iterable[str]
        Species order of split. Defaults to Stream.chemicals.IDs.
    solids : tuple[str]
        IDs of solids.
    moisture_content : float
        Fraction of water in stream.
    centrifuge_type : str
        Type of the centrifuge, either 'reciprocating_pusher' (1-20 ton/hr solids)
        or 'scroll_solid_bowl' (2-40 ton/hr solids).
    
    """
    _units = {'Solids loading': 'ton/hr',
              'Flow rate': 'm3/hr'}
    solids_loading_range = {
    'reciprocating_pusher': (1, 20),
    'scroll_solid_bowl': (2, 40)
    }
    kWhr_per_m3 = 1.40


    def _init(self, split, order=None, solids=None, moisture_content=0.40,
              centrifuge_type='scroll_solid_bowl', moisture_ID=None,
              strict_moisture_content=None):
        SolidsSeparator._init(
            self, moisture_content=moisture_content,
            split=split, order=order, moisture_ID=moisture_ID,
            strict_moisture_content=strict_moisture_content
        )
        if solids is None:
            solids = [i.ID for i in self.chemicals if i.locked_state == 's']
        self.solids = solids
        self.centrifuge_type = centrifuge_type
    
    @property
    def solids(self):
        return self._solids
    @solids.setter
    def solids(self, solids):
        self._solids = tuple(solids)
    
    @property
    def centrifuge_type(self):
        return self._centrifuge_type
    @centrifuge_type.setter
    def centrifuge_type(self, i):
        if not i in ('reciprocating_pusher', 'scroll_solid_bowl'):
            raise ValueError('`centrifuge_type` can only be "reciprocating_pusher" or '
                            f'"scroll_solid_bowl", not {i}.')
        self._centrifuge_type = i

    def _design(self):
        solids, centrifuge_type = self._solids, self.centrifuge_type
        ts = sum([s.imass[solids].sum() for s in self.ins if not s.isempty()]) # Total solids
        ts *= 0.0011023 # To short tons (2000 lbs/hr)
        self.design_results['Solids loading'] = ts
        lb, ub = self.solids_loading_range[centrifuge_type]
        if ts < lb:
            lb_warning(self, 'Solids loading', ts, 'ton/hr', lb)
        self.design_results['Number of centrifuges'] = ceil(ts/ub)
        cost = 68040*(ts**0.5) if centrifuge_type else 170100*(ts**0.3)
        cost *= bst.CE / 567
        self.baseline_purchase_costs['Centrifuges'] = cost
        self.F_BM['Centrifuges'] = 2.03
        self.design_results['Flow rate'] = F_vol_in = self.F_vol_in
        self.power_utility(F_vol_in * self.kWhr_per_m3)


class RotaryVacuumFilter(SolidsSeparator):
    """
    Create a RotaryVacuumFilter object.
    
    Parameters
    ----------
    ins : 
        * [0] Feed
        * [1] Wash water
    outs :  
        * [0] Retentate
        * [1] Permeate
    split : array_like or dict[str, float]
           Component splits.
    moisture_content : float
                       Fraction of water in retentate.
    
    """
    auxiliary_unit_names = ('vacuum_system',)
    _F_BM_default = {'Vessels': 2.32,
                     'Vacuum system': 1.0}
    
    #: Revolutions per second
    rps = 20/3600
    
    #: Radius of the vessel (m)
    radius = 1
    
    #: Suction pressure (Pa)
    P_suction = 1500.
    
    #: For crystals (lb/day-ft^2)
    filter_rate = 6000
    
    _kwargs = {'moisture_content': 0.80} # fraction
    _bounds = {'Individual area': (10, 800)}
    _units = {'Area': 'ft^2',
              'Individual area': 'ft^2'}
    
    def _design(self):
        flow = sum([stream.F_mass for stream in self.outs])
        self.design_results['Area'] = self._calc_Area(flow, self.filter_rate)
        
    def _cost(self):
        Design = self.design_results
        Area = Design['Area']
        ub = self._bounds['Individual area'][1]
        N_vessels = np.ceil(Area/ub)
        iArea = Area/N_vessels # individual vessel
        self.parallel['self'] = N_vessels
        self.parallel['vacuum_system'] = 1
        Design['Individual area'] = iArea
        self._load_vacuum_system(Area, N_vessels)
        logArea = np.log(iArea)
        Cost = np.exp(11.796 - 0.1905 * logArea + 0.0554 * logArea**2)
        self.baseline_purchase_costs['Vessels'] = Cost * bst.CE/567
    
    def _load_vacuum_system(self, area, N_vessels):
        s_cake, s_vacuumed = self.outs
        radius = self.radius
        Area = self.design_results['Individual area']
        N = self.parallel['self']
        vacummed_air = s_vacuumed.F_vol # Flow rate sucked-in displaces air
        air_density = 1.2754 # kg /m3
        self.vacuum_system = bst.VacuumSystem(
            F_mass=vacummed_air * air_density, 
            F_vol=vacummed_air, 
            P_suction=self.P_suction, 
            vessel_volume=N * radius * Area * 0.0929 / 2., # m3
        )
        
    @staticmethod
    def _calc_Area(flow, filter_rate):
        """Return area in ft^2 given flow in kg/hr and filter rate in lb/day-ft^2."""
        return flow * 52.91 / filter_rate
    
RVF = RotaryVacuumFilter

@cost('Flow rate', units='kg/hr', cost=1.5e6, CE=541.7,
      n=0.6, S=335e3, kW=2010, BM=2.3)
class CrushingMill(SolidsSeparator):
    """
    Create crushing mill unit operation for the 
    separation of sugarcane juice from the bagasse.
    
    Parameters
    ----------
    ins : 
        * [0] Shredded sugar cane
        * [1] Recycle water
    outs :  
        * [0] Bagasse
        * [1] Juice
    split : array_like or dict[str, float]
        Splits of chemicals to the bagasse.
    moisture_content : float
                       Fraction of water in Bagasse.
    
    """

_hp2kW = 0.7457
@cost('Retentate flow rate', 'Flitrate tank agitator',
      cost=26e3, CE=551, kW=7.5*_hp2kW, S=31815, n=0.5, BM=1.5)
@cost('Retentate flow rate', 'Discharge pump',
      cost=13040, CE=551, S=31815, n=0.8, BM=2.3)
@cost('Retentate flow rate', 'Filtrate tank',
      cost=103e3, S=31815, CE=551, BM=2.0, n=0.7)
@cost('Retentate flow rate', 'Feed pump', kW=74.57,
      cost= 18173, S=31815, CE=551, n=0.8, BM=2.3)
@cost('Retentate flow rate', 'Stillage tank 531',
      cost=174800, CE=551, S=31815, n=0.7, BM=2.0)
@cost('Retentate flow rate', 'Mafifold flush pump', kW=74.57,
      cost=17057, CE=551, S=31815, n=0.8, BM=2.3)
@cost('Retentate flow rate', 'Recycled water tank',
      cost=1520, CE=551, S=31815, n=0.7, BM=3.0)
@cost('Retentate flow rate', 'Wet cake screw',  kW=15*_hp2kW,
      cost=2e4, CE=521.9, S=28630, n=0.8, BM=1.7)
@cost('Retentate flow rate', 'Wet cake conveyor', kW=10*_hp2kW,
      cost=7e4, CE=521.9, S=28630, n=0.8, BM=1.7)
@cost('Retentate flow rate', 'Pressure filter',
      cost=3294700, CE=551, S=31815, n=0.8, BM=1.7)
@cost('Retentate flow rate', 'Pressing air compressor receiver tank',
      cost=8e3, CE=551, S=31815, n=0.7, BM=3.1)
@cost('Retentate flow rate', 'Cloth wash pump', kW=150*_hp2kW,
      cost=29154, CE=551, S=31815, n=0.8, BM=2.3)
@cost('Retentate flow rate', 'Dry air compressor receiver tank',
      cost=17e3, CE=551, S=31815, n=0.7, BM=3.1)
@cost('Retentate flow rate', 'Pressing air pressure filter',
      cost=75200, CE=521.9, S=31815, n=0.6, kW=112, BM=1.6)
@cost('Retentate flow rate', 'Dry air pressure filter (2)',
      cost=405000, CE=521.9, S=31815, n=0.6, kW=1044, BM=1.6)
class PressureFilter(SolidsSeparator):
    """
    Create a pressure filter for the separation of structural carbohydrates, 
    lignin, cell mass, and other solids. Capital costs are based on [2]_.
    
    Parameters
    ----------
    ins : 
        Contains structural carbohydrates, lignin, cell mass, and other solids.
    outs : 
        * [0] Retentate (i.e. solids)
        * [1] Filtrate
    split : array_like or dict[str, float]
        Splits of chemicals to the retentate. Defaults to values used in
        the 2011 NREL report on cellulosic ethanol as given in [2]_.
    moisture_content : float, optional
        Moisture content of retentate. Defaults to 0.35
    
    """
    _units = {'Retentate flow rate': 'kg/hr'}
    
    def _init(self, moisture_content=0.35, split=None):
        if split is None:
            chemicals = self.chemicals
            split = dict(
                Furfural=0.03571,
                Glycerol=0.03714,
                LacticAcid=0.03727,
                SuccinicAcid=0.03714,
                HNO3=0.03716,
                Denaturant=0.03714,
                DAP=0.03716,
                AmmoniumAcetate=0.03727,
                AmmoniumSulfate=0.03716,
                NaNO3=0.03716,
                Oil=0.03714,
                HMF=0.03571,
                Glucose=0.03647,
                Xylose=0.03766,
                Sucrose=0.0359,
                Mannose=0.0359,
                Galactose=0.0359,
                Arabinose=0.0359,
                Extract=0.03727,
                Tar=0.9799,
                CaO=0.9799,
                Ash=0.9799,
                NaOH=0.03716,
                Lignin=0.98,
                SolubleLignin=0.03727,
                GlucoseOligomer=0.03722,
                GalactoseOligomer=0.03722,
                MannoseOligomer=0.03722,
                XyloseOligomer=0.03722,
                ArabinoseOligomer=0.03722,
                Z_mobilis=0.9799,
                T_reesei=0.9799,
                Protein=0.98,
                Enzyme=0.98,
                Glucan=0.9801,
                Xylan=0.9811,
                Xylitol=0.03714,
                Cellobiose=0.0359,
                DenaturedEnzyme=0.98,
                Arabinan=0.9792,
                Mannan=0.9792,
                Galactan=0.9792,
                WWTsludge=0.9799,
                Cellulase=0.03727
            )
            remove_undefined_chemicals(split, chemicals)
            default_chemical_dict(split, chemicals, 0.03714, 0.03714, 0.9811)
        bst.SolidsSeparator._init(self, moisture_content=moisture_content, split=split)
    
    def _design(self):
        self.design_results['Retentate flow rate'] = self.outs[0].F_mass

PressureFilter._stacklevel += 1

#: TODO: Check BM assumption. Use 1.39 for crushing unit operations for now.
@cost('Flow rate', units='lb/hr', CE=567, lb=150, ub=12000, BM=1.39, 
      f=lambda S: exp((11.0991 - 0.3580*log(S) + 0.05853*log(S)**2)))
class ScrewPress(SolidsSeparator):
    """
    Create screw press unit operation for the 
    expression of liquids from solids. Capital cost is based on [1]_.
    
    Parameters
    ----------
    ins : 
        * [0] Solids + liquid
    outs :  
        * [1] Solids (retentate)
        * [0] Liquids (permeate)
    split : array_like or dict[str, float]
           Component splits.
    moisture_content : float
        Fraction of water in solids.
                  
    
    """ 
    kWh_per_bmt = 37.2 # From Perry's Handbook, 18-126
    # Energy consumption may be drastically different depending on the application
    # - 5 to 12 bdmt (tonne dry biomass) https://www.andritz.com/products-en/group/pulp-and-paper/service-solutions/screw-press-service/screw-press-upgrade-case-study-1-less
    
    def _cost(self):
        self._decorated_cost()
        biomass = self.ins[0]
        bmt = biomass.F_mass * 0.001
        self.add_power_utility(bmt * self.kWh_per_bmt)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from thermosteam import utils
from . import (patches,
               tictoc,
               piping,
               stream_link_options,
               functors,
               scope,
)
__all__ = ('colors',
           'patches', 
           'tictoc',
           'functors',
           *utils.__all__,
           *patches.__all__, 
           *tictoc.__all__, 
           *piping.__all__, 
           *stream_link_options.__all__,
           *functors.__all__,
           *scope.__all__,
)
from thermosteam.utils import *
from .patches import *
from .tictoc import *
from .piping import *
from .stream_link_options import *
from .functors import *
from .scope import *

del utils


# -*- coding: utf-8 -*-
"""
"""
import biosteam as bst
from numpy.testing import assert_allclose

def test_flash_energy_balance_with_pressure_drop():
    settings = bst.settings
    settings.set_thermo(['Water', 'Ethanol'], cache=True)
    settings.mixture.include_excess_energies = True
    feed = bst.Stream(None, Water=900, Ethanol=28000,
                      T=450, P=2000000)
    F1 = bst.Flash(None, ins=feed, P=101325, T=373.15) 
    F1.simulate()
    cooling, = F1.heat_utilities
    assert_allclose(F1.Hnet, cooling.unit_duty)
    
if __name__ == '__main__':
    test_flash_energy_balance_with_pressure_drop()


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from biorefineries.tests.test_biorefineries import (
    test_oilcane_O6, 
    test_oilcane_O8, 
    test_oilcane_O9,
    test_sugarcane,
    test_cornstover,
)
    
if __name__ == '__main__':
    test_oilcane_O6()
    test_oilcane_O8() 
    test_oilcane_O9()
    test_sugarcane()
    test_cornstover()


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
.. contents:: :local:

.. autoclass:: biosteam.units.adsorption.SingleComponentAdsorptionColumn

References
----------
.. [1] Adsorption basics Alan Gabelman (2017) Adsorption basics Part 1. AICHE

.. [2] Seader, J. D., Separation Process Principles: Chemical and Biochemical Operations,” 3rd ed., Wiley, Hoboken, NJ (2011).

.. [3] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani,
    R., & Ng, M. K. (2017). Product and Process Design Principles. Wiley.
    Cost Accounting and Capital Cost Estimation (Chapter 16)

"""
import biosteam as bst
from .design_tools import PressureVessel
from numpy.testing import assert_allclose
import numpy as np
from numba import njit
import flexsolve as flx
import matplotlib.pyplot as plt
from scipy.integrate import solve_ivp
from scipy.ndimage.filters import gaussian_filter
from thermosteam.units_of_measure import format_units

__all__ = ('SingleComponentAdsorptionColumn', 'AdsorptionColumn',)

@njit(cache=True)
def fixed_step_odeint(f, x0, step, tf, args):
    M = int(tf/step)
    N = len(x0)
    ts = np.linspace(0, tf, M)
    xs = np.zeros((M, N))
    xs[0] = x0
    for i in range(1, M):
        dx_dt = f(ts[i], x0, *args)
        x0 += dx_dt * step
        xs[i] = x0 
    return (ts, xs)

@njit(cache=True)
def equilibrium_loading_Langmuir_isotherm(
        C,  # Solute concentration in the feed [g / L]
        KL,  # Equilibrium constant
        q_max,  # Maximum equilibrium loading mg/g
    ):
    return C * KL * q_max / (1 + KL * C)  # g / kg

@njit(cache=True)
def equilibrium_loading_Freundlich_isotherm(
        C,  # Solute concentration in the feed [g / L]
        KL,  
        n,  
    ):
    return KL * C**(1/n) # g / kg

@njit(cache=True)
def estimate_equilibrium_bed_length(
        C,  # Solute concentration in the feed [kg / m3]
        u,  # Superficial velocity [m / h]
        cycle_time,  # Cycle time [h]
        rho_adsorbent,  # Bulk density of the bed [kg / m3]
        q0,  # Equilibrium loading [g / kg]
    ):
    q_capacity = q0 * rho_adsorbent / 1000 # kg / m3
    L = C * u * cycle_time / q_capacity  # m
    return L

@njit(cache=True)
def dCdt_optimized_Langmuir(
        t, C, N_slices, 
        Da, dz, KL_qm,
        q0_over_C0,
        q0_KL,
        beta_q0_rho_over_C0,
    ):
    CL = C[:N_slices] 
    qt = C[N_slices:] # q-avg
    CL_ = CL.copy()
    CL_[CL < 0] = 0
    qe = (CL_ * KL_qm)/(q0_over_C0 + q0_KL * CL_)
    dCL_dz = CL_
    dCL_dz[1:] = (CL[1:] - CL[:-1]) / dz
    dCL_dz[0] = (CL[0] - 1) / dz
    dC_dt = C.copy()
    dq_dt = Da * (qe - qt)
    dCL_dt = -dCL_dz - beta_q0_rho_over_C0 * dq_dt
    dC_dt[:N_slices] = dCL_dt
    dC_dt[N_slices:] = dq_dt
    return dC_dt 

@njit(cache=True)
def dCdt(
        t, C, N_slices, 
        Da, dz, isotherm_model, isotherm_args,
        beta_q0_rho_over_C0, C0, q0
    ):
    CL = C[:N_slices] 
    qt = C[N_slices:] # q-avg
    CL_ = CL * C0
    mask = CL_ < 0
    CL_[mask] = 0
    qe = isotherm_model(CL_, *isotherm_args) / q0
    qe[qe > q0] = q0
    dCL_dz = CL.copy()
    dCL_dz[1:] = (CL[1:] - CL[:-1]) / dz
    dCL_dz[0] = (CL[0] - 1) / dz
    dC_dt = C.copy()
    dq_dt = Da * (qe - qt)
    dCL_dt = -dCL_dz - beta_q0_rho_over_C0 * dq_dt
    dC_dt[:N_slices] = dCL_dt
    dC_dt[N_slices:] = dq_dt
    return dC_dt 

def adsorption_bed_pressure_drop(
        D = 0.001, # particle diameter [m]
        rho = 800, # liquid density [kg/m3]
        mu = .001, # viscosity [kg/m*s]
        epsilon = 1 - 0.38/0.8, # void fraction
        u = 14/3600, # superficial velosity [m/s]
        L = 5, # length of bed [m]
    ):
    re = 1 - epsilon
    e2 = epsilon * epsilon
    e3 = e2 * epsilon
    N_Re = D * u * rho/(mu*re)
    if N_Re < 10: # Kozeny-Carman equation
        dP = 150 * mu * u * re ** 2 * L/(D ** 2 * e3)
    elif 10 < N_Re < 1000: # Ergun equation 
        dP = 1.75 * u**2 * L * re * rho/(D * e3) + 150 * mu * u * re**2/(D**2 * e3)
    else: # Burke-Plummer equation
        dP = 1.75 * u**2 * L * re * rho/(D * e3) + 150
    return dP

class SingleComponentAdsorptionColumn(PressureVessel, bst.Unit):
    """
    Create an adsorption column which can operate by temperature or pressure swing,
    or by disposing the adsorbent. Heats of adsorption are neglected but may become
    an option with future development in BioSTEAM. Default parameters are heuristic values 
    for adsorption of water and polar components using activated carbon. The
    design and cost algorithm of the adsorption column is based on [1]_, [2]_, 
    and [3]_. 
    
    By default, three vessels are used: a lead, a guard, and a standby vessel. 
    The lead is in equilibrium, guard holds the breakthrough curve and mass 
    transfer zone, and the standby is in regeneration. Once the standby vessel
    is regenerated, it becomes the guard, the lead becomes the standby, and 
    the guard becomes the lead. Therefore, the cycle time is the regeneration
    time. 
    
    For temperature or pressure swing adsorption, set the temperature or pressure
    of the regeneration fluid. If a regeneration fluid is used, the flow rate 
    of regeneration fluid solved for such that it meets the required cycle time.
    
    Parameters
    ----------
    ins : 
        * [0] Feed
        * [1] Regeneration fluid
        * [2] Adsorbent (if no regeneration fluid is used).
    outs : 
        * [0] Effluent
        * [1] Purge
        * [2] Spent adsorbent (if no regeneration fluid is used)
    k : float, optional
        Mass transfer coefficient [1/h]. If not given, mass tranfer model is 
        ignored.
    superficial_velocity : float, optional
        Superficial velocity of the feed. The diameter of the receiving vessel adjusts
        accordingly. Defaults to 7.2 [m / hr]. Typical velocities are 4 to 14.4 m / hr for liquids [1]_.
        Common velocity range for gasses is 504-2160 m / hr [1]_.
    cycle_time : float, optional
        Time at which the receiving vessel is switched. Defaults to 3 [hr]. 
        Note that 1-2 hours required for thermal-swing-adsorption 
        (TSA) for silica gels [2]_. 
    isotherm_args : tuple[float, ...], optional
        Arguments to pass to the isotherm model. 
        If isotherm model is 'Langmuir', arguments must be:
        * KL: Equilibrium constant [L/mg] for Langmuir isotherm
        * q_max: Maximum equilibrium loading [mg/g] for Langmuir isotherm
        If isotherm model is 'Freundlich', arguments must be:
        * K: Equilibrium constant [L/mg] for Freundlich isotherm
        * n: exponential coefficient for Freundlich isotherm
    isotherm_model : Callable|str, optional, 
        Can be 'Langmuir', 'Freundlich', or a function.
        If a function is given, it should be in the form of f(C, *args) -> g absorbate / kg absorbent.
    void_fraction : 0.525 
        Fraction of empty space in the adsorbent by vol.
    rho_adsorbent : float, optional
        The density of the adsorbent. Defaults to 380 [kg/m3], which is common 
        for activated carbon granules.
    regeneration_fluid : dict[str, float]
        Arguments to initialize fluid used to regenerate the bed.
    LUB : float, optional 
        Length of unused bed. Defaults to 1.219 m if no mass transfer 
        coefficient is given. This parameter is ignored if isotherms are provided.
    P : float, optional
        Operating pressure of the column. Defaults to 101325.
    C_final_scaled : float, optional
        Final outlet concentration at breakthrough relative to inlet. 
        Defaults to 0.05.
    regeneration_fluid : dict[str, float]
        Regeneration fluid composition and thermal conditions.
    k_regeneration : float, optional 
        Mass transfer coefficient of the regeration solvent [1/h].
    regeneration_isotherm_args : tuple[float, ...], optional
        Arguments to regeneration isotherm model.
    regeneration_isotherm_model : Callable|str
        Isotherm model for regenerating the adsorbent.
    N_slices : int, optional
        Number of slices to model mass transfer. Defaults to 80.
    adsorbate : str
        Name of adsorbate.
    vessel_material : float, optional
        Vessel material. Defaults to 'Stainless steel 316',
    vessel_type : float, optional
        Vessel type. Defaults to 'Vertical'.
    adsorbent : str
        Name of adsorbent. Defaults to 'ActivatedCarbon'.
    N_columns : int
        Number of columns can be either 2 or 3. The 3-column configuration 
        minimizes cost of adsorbent regeneration. The 2-column configuration 
        minimizes capital cost.
    particle_diameter : float
        Particle diameter assumed for pressure drop estimation [m]. 
        Defaults to 0.004 m.    
    
    Examples
    --------
    >>> import biosteam as bst
    >>> bst.settings.set_thermo([
    ...    'Water', 
    ...    bst.Chemical('Adsorbate', search_db=False, default=True, phase='l'),
    ...    bst.Chemical('ActivatedCarbon', search_db=False, default=True, phase='s')
    ... ])
    >>> feed = bst.Stream(ID='feed', phase='l', T=298, P=1.01e+06,
    ...                   Water=1000, Adsorbate=0.001, units='kg/hr')
    >>> A1 = bst.AdsorptionColumn(
    ...     'A1', ins=feed, outs='effluent',
    ...     cycle_time=1000,
    ...     superficial_velocity=9.2,
    ...     isotherm_model='Langmuir',
    ...     isotherm_args=(1e3, 7.), # K [g / L], q_max [g / kg]
    ...     k=0.3, # [1 / hr]
    ...     void_fraction=0.525, 
    ...     C_final_scaled=0.05,
    ...     adsorbate='Adsorbate',
    ...     particle_diameter=0.004,
    ... )
    >>> A1.simulate()
    >>> # A1.simulation_gif() # Create a gif of the fluid concentration profile over time.
    
    .. image:: ../../images/adsorption_column.gif
       :align: center
       :alt: Adsorption column fluid concentration profile over time.
    
    >>> A1.results()
    Single component adsorption column                  Units                   A1
    Electricity         Power                              kW                0.241
                        Cost                           USD/hr               0.0188
    Design              Diameter                           ft                 1.22
                        Length                             ft                 12.2
                        Vessel type                                       Vertical
                        Weight                             lb                  521
                        Wall thickness                     in                 0.25
                        Pressure drop                      Pa                  103
                        Vessel material                        Stainless steel 316
    Purchase cost       Vertical pressure vessel (x3)     USD             6.12e+04
                        Platform and ladders (x3)         USD             8.37e+03
                        Pump - Pump (x3)                  USD             1.31e+04
                        Pump - Motor (x3)                 USD                  174
    Total purchase cost                                   USD             8.28e+04
    Utility cost                                       USD/hr               0.0188
    
    >>> A1.show('wt')
    SingleComponentAdsorptionColumn: A1
    ins...
    [0] feed  
        phase: 'l', T: 298 K, P: 1.01e+06 Pa
        flow (kg/hr): Water      1e+03
                      Adsorbate  0.001
    [1] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow: 0
    [2] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kg/hr): ActivatedCarbon  0.154
    outs...
    [0] effluent  
        phase: 'l', T: 298 K, P: 101325 Pa
        flow (kg/hr): Water  1e+03
    [1] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow: 0
    [2] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kg/hr): ActivatedCarbon  0.154
    
    """
    auxiliary_unit_names = (
        'pump', 'heat_exchanger',
        'regeneration_pump'
    )
    _N_ins = 3
    _N_outs = 3

    _units = {'Pressure drop': 'Pa',
              **PressureVessel._units}

    # Cost of regeneration in $/m3
    adsorbent_cost = {
        'ActivatedAlumina': 72 * 0.0283168,
        'ActivatedCarbon': 41 * 0.0283168,
        'SilicaGel': 210 * 0.0283168,
        'MolecularSieves': 85 * 0.0283168,
    }

    # TODO: Update later with reference
    # This is only active if modeling regeneration of solute.
    _default_equipment_lifetime = {
        'ActivatedAlumina': 100,
        'ActivatedCarbon': 100,
        'SilicaGel': 100,
        'MolecularSieves': 100,
    }

    isotherm_models = {
        'Langmuir': equilibrium_loading_Langmuir_isotherm,
        'Freundlich': equilibrium_loading_Freundlich_isotherm,
    }

    def _init(self,
            cycle_time,
            # If not given, mass tranfer model is ignored
            k=None,  # Mass transfer coefficient [1/h]
            # Langmuir:
            # - KL: Equilibrium constant [L/mg] for Langmuir isotherm
            # - q_max: Maximum equilibrium loading [mg/g] for Langmuir isotherm
            # Freundlich:
            # - K: Equilibrium constant [L/mg] for Freundlich isotherm
            # - n: exponential coefficient for Freundlich isotherm
            isotherm_args=None,
            # Can be a function, 'Langmuir', or 'Freundlich'.
            # If a function is given, it should be in the form of f(C, *args) -> g absorbate / kg absorbent
            isotherm_model=None, 
            k_regeneration=None, # Mass transfer coefficient of the regeration solvent [1/h]
            regeneration_isotherm_args=None,
            regeneration_isotherm_model=None,
            # Note that the columns are sized according to the limiting isotherm.
            LUB_forced=None, # Defaults to 0.6096 m if no mass transfer coefficient is given.
            void_fraction=1 - 0.38 / 0.8, # Solid by vol [%]
            rho_adsorbent=380, # kg/m3
            superficial_velocity=14.4,  # [m / h]
            P=101325,
            C_final_scaled=0.05, # Final outlet concentration at breakthrough relative to inlet.
            N_slices=int(50), # Number of slices to model mass transfer.
            regeneration_fluid=None, # [dict] Regeneration fluid composition and thermal conditions.
            adsorbate=None, # [str] Name of adsorbate.
            vessel_material='Stainless steel 316',
            vessel_type='Vertical',
            adsorbent='ActivatedCarbon',
            N_columns=3, # 3 column configuration minimizes cost of adsorbent regeneration. 2 column configuration minimizes capital cost.
            particle_diameter = 0.004, # [m]
        ):
        if N_columns not in (2, 3):
            raise ValueError('only 2 or 3-column configurations are valid')
        if isotherm_model is None:
            if isotherm_args is not None:
                raise ValueError('no isotherm model given')
            isotherm_model = lambda C: 0.1
        elif isinstance(isotherm_model, str):
            if isotherm_model in self.isotherm_models:
                isotherm_model = self.isotherm_models[isotherm_model]
            else:
                raise ValueError(
                    f'isotherm model {isotherm_model} is not available; '
                    f'valid options are {list(self.isotherm_models)}'
                )
        if regeneration_isotherm_model is None:
            if regeneration_isotherm_args is not None:
                raise ValueError('no regeneration isotherm model given')
        elif isinstance(isotherm_model, str):
            if regeneration_isotherm_model in self.isotherm_models:
                regeneration_isotherm_model = self.isotherm_models[regeneration_isotherm_model]
            else:
                raise ValueError(
                    f'isotherm model {regeneration_isotherm_model} is not available; '
                    f'valid options are {list(self.isotherm_models)}'
                )
        if k is None and k_regeneration is None:
            if LUB_forced is None: LUB_forced = 0.6096
        elif LUB_forced is not None:
            raise ValueError('length of unused bed given, but will be '
                             'estimated based on mass transfer modeling')
        if regeneration_isotherm_args is None: regeneration_isotherm_args = ()
        if isotherm_args is None: isotherm_args = ()
        self.adsorbate = adsorbate
        self.regeneration_fluid = regeneration_fluid
        self.N_columns = N_columns
        self.cycle_time = cycle_time
        self.superficial_velocity = superficial_velocity
        self.P = P
        self.isotherm_args = isotherm_args
        self.isotherm_model = isotherm_model
        self.regeneration_isotherm_args = regeneration_isotherm_args
        self.regeneration_isotherm_model = regeneration_isotherm_model
        self.k = k
        self.k_regeneration = k_regeneration
        self.void_fraction = void_fraction
        self.rho_adsorbent = rho_adsorbent
        self.C_final_scaled = C_final_scaled
        self.vessel_material = vessel_material
        self.vessel_type = vessel_type
        self.N_slices = N_slices
        self.particle_diameter = particle_diameter
        self.adsorbent = adsorbent
        self.LUB_forced = LUB_forced
        self.auxiliary('pump', bst.Pump, ins=self.ins[0])
        if regeneration_fluid:
            regeneration_pump = self.auxiliary('regeneration_pump', bst.Pump, ins=self.ins[1])
            self.auxiliary('heat_exchanger', bst.HXutility, ins=regeneration_pump.outs[0])

    def _run(self):
        feed, regeneration_fluid, adsorbent = self.ins
        outlet, spent_fluid, spent_adsorbent = self.outs
        regeneration_fluid.empty()
        spent_fluid.empty()
        adsorbent.empty()
        spent_adsorbent.empty()
        outlet.copy_like(feed)
        outlet.P = self.P
        outlet.imol[self.adsorbate] = 0
        if self.regeneration_fluid:
            regeneration_fluid = self.heat_exchanger.outs[0]
            regeneration_fluid.reset(**self.regeneration_fluid)
            self.heat_exchanger.T = regeneration_fluid.T
            self._size_columns()
            self.regeneration_superficial_velocity = u = self._solve_regeneration_velocity()
            Q = self.area * u
            regeneration_fluid.F_vol = Q * 3600 # m3/s to m3/h
            self.heat_exchanger.ins[0].copy_flow(regeneration_fluid)
            self.heat_exchanger.run()
            spent_fluid.copy(regeneration_fluid)
            spent_fluid.imol[self.adsorbate] = feed.imol[self.adsorbate]

    def column_widget(self):
        # Not currently working
        import ipywidgets as widgets
        from ipywidgets import interact
        t = self.t
        N_time = len(t)
        tf = t[-1]
        z = np.arange(0, 1, 1/self.N_slices)
        CL_scaled = self.CL_scaled
        plt.xlabel('z')
        plt.yticks([0, 0.2, 0.4, 0.6, 0.8, 1])
        plt.ylabel('Concentration')
        def plot(time):
            time_index = int(time / tf * N_time) - 1
            plt.clf()
            plt.plot(z, CL_scaled[:, time_index], label = '$\theta$ vs z')
            plt.show()
        
        slider = widgets.FloatSlider(min=0, max=tf, step=tf/100, value=0.5*tf)
        return interact(plot, time=slider)

    def simulation_gif(self, liquid=True):
        import matplotlib.animation as animation
        from scipy.interpolate import interp1d
        t = self.t_scaled
        z = np.arange(0, 1, 1/self.N_slices)
        y = self.CL_scaled if liquid else self.q_scaled
        fig = plt.figure()
        N_frames = 120
        f = interp1d(t, y)
        x = np.linspace(0, t[-1], N_frames)
        y = f(x)
        def animate(frame):
            plt.clf()
            artist = plt.plot(z, y[:, frame])
            plt.xlabel(r'z = x / L')
            plt.ylabel(r'$\theta$ = C / C$_{feed}$')
            plt.yticks([0, 0.2, 0.4, 0.6, 0.8, 1])
            plt.ylim([0, 1])
            return (artist,)
        
        ani = animation.FuncAnimation(
            fig, animate, repeat=True,
            frames=N_frames - 1, interval=66,
        )
        
        # To save the animation using Pillow as a gif
        writer = animation.PillowWriter(fps=15,
                                        metadata=dict(artist='Me'),
                                        bitrate=1800)
        ani.save('adsorption_column.gif', writer=writer)
        return ani

    def _simulate_adsorption_bed(
            self,
            regeneration,
            L, # m
            u, # [m / hr]
        ):
        cycle_time = self.cycle_time
        N_slices = self.N_slices
        C_init = np.zeros(2 * N_slices)
        void_fraction = self.void_fraction
        rho_adsorbent = self.rho_adsorbent
        if regeneration:
            C_init[N_slices:] = 1 # saturated column for regeneration
            isotherm_model = self.regeneration_isotherm_model
            isotherm_args = self.regeneration_isotherm_args
            k = self.k_regeneration
        else:
            k = self.k
            isotherm_model = self.isotherm_model
            isotherm_args = self.isotherm_args
        Da = k * L / u
        beta = (1 - void_fraction)/void_fraction
        self.dz = dz = 1 / (N_slices - 1)
        t_scale = (L / u)
        C0 = self.C0
        q0 = isotherm_model(C0, *isotherm_args)
        beta_q0_rho_over_C0 = (beta * q0 * rho_adsorbent / 1000) / C0
        if isotherm_model is equilibrium_loading_Langmuir_isotherm:
            KL, q_m = isotherm_args
            KL_qm = KL * q_m
            q0_over_C0 = q0 / C0
            q0_KL = q0 * KL
            args = (N_slices, Da, dz, KL_qm,
                    q0_over_C0, q0_KL, beta_q0_rho_over_C0)
            f = dCdt_optimized_Langmuir
            sol = solve_ivp(
                f, t_span=(0, cycle_time / t_scale), 
                y0=C_init, 
                method='BDF',
                args=args,
            ) 
            CL = sol.y[:N_slices]
            q = sol.y[N_slices:]
            t = sol.t
        else:
            args = (N_slices, Da, dz, isotherm_model, isotherm_args,
                    beta_q0_rho_over_C0, C0, q0)
            f = dCdt
            time_step = 0.2 / N_slices
            t, Y = fixed_step_odeint(
                f, C_init, time_step, cycle_time / t_scale, args
            )
            t = np.array(t)
            Y = gaussian_filter(np.array(Y).T, 5, axes=1)
            CL = Y[:N_slices]
            q = Y[N_slices:]
        if regeneration:
            self.rt_scaled = t
            self.rCL_scaled = CL
            self.rq_scaled = q
            self.rt_scale = t_scale
        else:
            self.t_scaled = t
            self.CL_scaled = CL
            self.q_scaled = q
            self.t_scale = t_scale
    
    def _estimate_length_of_unused_bed(self, LUB):
        L_guess = self.LES + LUB
        self._simulate_adsorption_bed(
            False, L_guess, self.superficial_velocity, 
        )
        CL = self.CL_scaled
        C_min = self.C_final_scaled
        C_max = 1 - C_min
        mask = (CL > C_max).any(axis=0) & (CL < C_min).any(axis=0) # slices that have breakthrough curve
        time_index = np.where(mask)[0][-1] # As close to breakthrough as possible
        start_index = np.where(CL[:, time_index] >= C_max)[0][-1]
        end_index = np.where(CL[:, time_index] <= C_min)[0][0]
        length = end_index - start_index
        self.t_scaled = self.t_scaled[:time_index]
        self.CL_scaled = self.CL_scaled[:, :time_index]
        self.q_scaled = self.q_scaled[:, :time_index]
        self.MTZ = length * self.dz * L_guess
        self.LUB = LUB = self.MTZ / 2
        return LUB
    
    def _regeneration_time_objective(self, 
            u, # [m / hr]
        ):
        self._simulate_adsorption_bed(
            True, self.column_length, u,
        )
        q = self.q_scaled
        mask = (q > 1e-6).all(axis=0) & (q < 1e-3).all(axis=0) # slices that have finished regenerating
        time_index = np.where(mask)[0][-1] # Conservative time point
        return self.t_scaled[time_index] * self.t_scale - self.cycle_time
    
    def _solve_regeneration_velocity(self):
        self.regeneration_velocity = u = flx.IQ_interpolation(
            self._estimate_regeneration_time, 4, 14.4, self.cycle_time, 
            xtol=1e-3, ytol=1e-2, 
        ) 
        return u
        
    # def _final_concentration_objective(self, 
    #         L_guess, # m
    #         u, # [m / hr]
    #         cycle_time, # [h]
    #         void_fraction, # external void fraction
    #         q_m, # [g / kg]
    #         k, # [1 / hr]
    #         KL, # [L / g]
    #         feed_rho, # [kg / L]     
    #         C0, # [g / L]
    #     ):
    #     self._simulate_adsorption_bed(
    #         L_guess, u, cycle_time, void_fraction, q_m, k, KL, feed_rho, C0, 
    #     )
    #     C_final = self.CL_scaled[-1, -1] # Final concentration at the end of the bed after cycle time
    #     return C_final - self.C_final
    
    # def _solve_length_of_unused_bed(self, 
    #         u, # [m / hr]
    #         cycle_time, # [h]
    #         void_fraction, # external void fraction
    #         q_m, # [g / kg]
    #         k, # [1 / hr]
    #         KL, # [L / g]
    #         feed_rho, # [kg / L]     
    #         C0, # [g / L]
    #     ):
        
    #     L_guess = self.LES + getattr(self, 'LUB', 2/3) # # 2/3 is a heuristic from AICHE adsorption basics part 1
    #     L = secant(
    #         self._final_concentration_objective, L_guess, 
    #         xtol=self.LES/100, ytol=self.C_final/100,
    #         args=(u, cycle_time, void_fraction, q_m, k, KL, feed_rho, C0)
    #     ) 
    #     LUB = L - self.LES
    #     assert LUB > 0
    #     return LUB

    def estimate_length_of_unused_bed(self):
        LUB_guess = getattr(self, 'LUB', 2/3) # 2/3 is a heuristic from AICHE adsorption basics part 1
        return flx.wegstein(
            self._estimate_length_of_unused_bed, LUB_guess,
            xtol=self.LES * 1e-2, maxiter=3, checkiter=False,
        )

    def _size_columns(self):
        # Design based on capacity (as estimated by the Langmuir isotherm).
        # The length of the mass transfer zone (MTZ) is assumed to be 4 ft based
        # on AICHE adsorption basics part 1. However, future work will use the rate
        # constant and mass transfer modeling to estimate the length of the MTZ.
        feed = self.ins[0]
        Q = feed.F_vol  # m3 / hr
        u = self.superficial_velocity
        self.area = A = Q / u
        self.diameter = diameter = 2 * (A/np.pi) ** 0.5
        self.C0 = C_feed = feed.imass[self.adsorbate] / Q  # g/L or kg/m3
        if C_feed == 0: 
            self.q0 = self.LES = self.LUB = self.total_length = self.column_length = 0
            return 0, 0
        cycle_time = self.cycle_time
        self.q0 = q0 = self.isotherm_model(
            C_feed, *self.isotherm_args
        )
        self.LES = LES = estimate_equilibrium_bed_length(
            C_feed, u, cycle_time, self.rho_adsorbent, q0
        )
        if self.LUB_forced is None:
            self.LUB = LUB = self.estimate_length_of_unused_bed()
        else:
            self.LUB = LUB = self.LUB_forced
        self.total_length = total_length = LES + LUB
        if self.N_columns == 3: 
            column_length = total_length / 2
        elif self.N_columns == 2:
            column_length = total_length
        self.column_length = column_length
        return diameter, column_length

    def _design(self):
        diameter, column_length = self._size_columns()
        feed = self.ins[0]
        if column_length == 0:
            self.design_results['Weight'] = 0
            self.design_results['Diameter'] = 0
            self.design_results['length'] = 0
            return
        self.set_design_result('Diameter', 'm', diameter)
        self.set_design_result('Length', 'm', column_length)
        self.design_results.update(
            self._vertical_vessel_design(
                feed.get_property('P', 'psi'),
                self.design_results['Diameter'],
                self.design_results['Length']
            )
        )
        dP = adsorption_bed_pressure_drop(
            D = self.particle_diameter,
            rho = feed.rho,
            mu = feed.get_property('mu', 'kg/m/s'), # viscosity [kg/m*s]
            epsilon = self.void_fraction, # void fraction
            u = self.superficial_velocity / 3600, # superficial velosity [m/s]
            L = self.total_length, # length of bed [m]
        )
        self.set_design_result('Pressure drop', 'Pa', dP)
        self.pump.P = (self.P - feed.P) + dP
        self.pump.simulate()
        if self.regeneration_fluid:
            feed = self.ins[1]
            outlet = self.outs[1]
            self.regeneration_pump.P = (outlet.P - feed.P) + adsorption_bed_pressure_drop(
                D = self.particle_diameter,
                rho = feed.rho,
                mu = feed.get_property('mu', 'kg/m/s'), # viscosity [kg/m*s]
                epsilon = self.void_fraction, # void fraction
                u = self.regeneration_superficial_velocity / 3600, # superficial velosity [m/s]
                L = self.total_length, # length of bed [m]
            )

    def _cost(self):
        design_results = self.design_results
        weight = design_results['Weight']
        baseline_purchase_costs = self.baseline_purchase_costs
        if weight == 0:
            baseline_purchase_costs.clear()
            self.parallel.clear()
            return
        baseline_purchase_costs.update(
            self._vessel_purchase_cost(
                weight, design_results['Diameter'], design_results['Length']
            )
        )
        self.parallel['self'] = self.N_columns
        if self.regeneration_fluid:
            baseline_purchase_costs[self.adsorbent] = (
                self.N_columns * self.area * self.column_length
                * self.adsorbent_cost[self.adsorbent]
            )
        else:
            self.ins[2].imass[self.adsorbent] = self.outs[2].imass[self.adsorbent] = self.area * self.column_length / self.cycle_time * self.rho_adsorbent
            self.ins[2].price = self.adsorbent_cost[self.adsorbent] / self.rho_adsorbent

    @staticmethod
    def fit_Freundlich_isotherm(C, q):
        y = np.log(q)
        x = np.log(C)
        m, b = np.polyfit(x, y, 1)
        n = 1 / m
        K = np.exp(b)
        y_ = m * x + b
        q_ = equilibrium_loading_Freundlich_isotherm(C, K, n)
        R2 = np.corrcoef(y, y_)**2
        R2q = np.corrcoef(q, q_)**2
        return {
            'K': K,
            'n': n,
            'parameters': (K, n),
            'linear coefficients': (m, b),
            'linearized data': (x, y),
            'linearized prediction': (x, y_),
            'R2': R2[0, 1],
            'R2-q': R2q[0, 1],
        }
    
    @staticmethod
    def fit_Langmuir_isotherm(C, q):
        y = C / q
        x = C
        m, b = np.polyfit(x, y, 1)
        qmax = 1 / m
        KL = 1 / (b * qmax)
        y_ = m * x + b
        q_ = equilibrium_loading_Langmuir_isotherm(C, KL, qmax)
        R2q = np.corrcoef(q, q_)**2
        R2 = np.corrcoef(y, y_)**2
        return {
            'K': KL,
            'qmax': qmax,
            'parameters': (KL, qmax),
            'linear coefficients': (m, b),
            'linearized data': (x, y),
            'linearized prediction': (x, y_),
            'R2': R2[0, 1],
            'R2-q': R2q[0, 1],
        }
    
    @classmethod
    def plot_isotherm_fit(cls, C, q, method):
        if method == 'Langmuir':
            dct = cls.fit_Langmuir_isotherm(C, q)
        elif method == 'Freundlich':
            dct = cls.fit_Freundlich_isotherm(C, q)
        else:
            raise ValueError("method must be either 'Langmuir' or 'Freundlich'")
        plt.scatter(*dct['linearized data'])
        R2 = dct['R2']
        plt.plot(*dct['linearized prediction'],
                 label=f'R$^2$ = {np.round(R2, 2)}')
        if method == 'Langmuir':
            plt.xlabel(f'C')
            plt.ylabel(f'C / q')
        else:
            plt.xlabel(f'log(C)')
            plt.ylabel(f'log(q)')
    
    @staticmethod
    def fit_solid_mass_transfer_coefficient(t, C, volume, adsorbent, model, args):
        C0 = C[0]
        q = volume * (C0 - C) / adsorbent
        qe = model(C, *args)
        y = np.log((qe - q) / qe)
        x = t
        m = (x * y).sum() / (x * x).sum()
        y_ = m * x
        R2 = np.corrcoef(y, y_)**2
        return {
            'k': -m,
            'linearized data': (x, y),
            'linearized prediction': (x, y_),
            'R2': R2[0, 1],
        }
    
    @staticmethod
    def plot_isotherm_and_mass_transfer_coefficient_fit(
            Ce, qe, t, C, volume, adsorbent, method=None, 
            C_units=None, q_units=None, t_units=None,
        ):
        if C_units is None: C_units = ''
        else: C_units = f' [{format_units(C_units)}]'
        if q_units is None: q_units = ''
        else: q_units = f' [{format_units(q_units)}]'
        if t_units is None: t_units = ''
        else: t_units = f' [{format_units(t_units)}]'
        def plot_fit_isotherm(dct, method):
            plt.scatter(*dct['linearized data'])
            if method == 'Langmuir':
                plt.plot(
                    *dct['linearized prediction'],
                    label=f"R$^2$ = {dct['R2']:.3g}, K={dct['K']:.3g}, q$_{{max}}$={dct['qmax']:.3g}"
                )
                xlabel = 'C {}'
                ylabel = 'C / q {}'
            else:
                plt.plot(
                    *dct['linearized prediction'],
                    label=f"R$^2$ = {dct['R2']:.3g}, K={dct['K']:.3g}, n={dct['n']:.3g}"
                )
                xlabel = 'log(C{})'.format(C_units)
                ylabel = 'log(q{})'.format(q_units)
            plt.xlabel(xlabel)
            plt.ylabel(ylabel)
            plt.legend()
            return dct
        dct_L = bst.AdsorptionColumn.fit_Langmuir_isotherm(Ce, qe)
        dct_F = bst.AdsorptionColumn.fit_Freundlich_isotherm(Ce, qe)
        if method == 'Langmuir' or (method is None and dct_L['R2-q'] > dct_F['R2-q']):
            isotherm_model = 'Langmuir'
            dct = dct_L
        else:
            isotherm_model = 'Freundlich'
            dct = dct_F
        fig, axes = plt.subplots(2, 1)
        ax1, ax2 = axes
        plt.sca(ax1)
        plot_fit_isotherm(dct, isotherm_model)
        model = bst.AdsorptionColumn.isotherm_models[isotherm_model]
        args = dct['parameters']
        dct_k = bst.AdsorptionColumn.fit_solid_mass_transfer_coefficient(
            t, C, volume, adsorbent, model, args
        )
        plt.sca(ax2)
        plt.scatter(*dct_k['linearized data'])
        R2 = dct_k['R2']
        plt.plot(
            *dct_k['linearized prediction'],
            label=f"R$^2$ = {R2:.2g}, k={dct_k['k']:.3g}"
        )
        plt.ylabel('log((qe - q) / q)')
        plt.xlabel(f't{t_units}')
        plt.legend()
        plt.subplots_adjust(hspace=0.3, wspace=0.3)
        return fig, axes, {
            'isotherm_model': isotherm_model, 
            'isotherm_args': args,
            'k': dct_k['k'],
        }

AdsorptionColumn = SingleComponentAdsorptionColumn



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2024, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from __future__ import annotations
import pandas as pd
from warnings import warn
from thermosteam._graphics import UnitGraphics
from ._heat_utility import HeatUtility
from .utils import AbstractMethod, format_title, static, piping
from ._power_utility import PowerUtility
from .exceptions import UnitInheritanceError
from thermosteam.units_of_measure import convert
from copy import copy
import biosteam as bst
from thermosteam import Stream, AbstractUnit, ProcessSpecification
from numpy.typing import NDArray
from typing import Callable, Optional, TYPE_CHECKING, Sequence, Iterable
from thermosteam.base.sparse import SparseVector, sum_sparse_vectors
import numpy as np
import thermosteam as tmo
from thermosteam import EquationNode
if TYPE_CHECKING: 
    System = bst._recycle_system
    HXutility = bst.HXutility
    UtilityAgent = bst.UtilityAgent

streams = Optional[Sequence[Stream|str]]

__all__ = ('Unit',)

# %% Typing

# from typing import Collection, Union, Annotated
# streams = Union[Collection[Union[Stream, str, None]], Union[Stream, str, None]]
# stream = Union[Annotated[Union[Stream, str, None], 1], Union[Stream, str, None]]
# stream_sequence = Collection[Union[Stream, str, None]]

# %% Unit Operation

def phenomena_oriented_run(self):
    if not (self._recycle_system and self._system.algorithm == 'Phenomena oriented'):
        Unit.run(self)
        return
    ins = self.ins
    outs = self.outs
    Ts = [i.T for i in outs]
    Q = self.Hnet
    Unit.run(self)
    new = sum(
        [i.mol for i in outs],
        -sum([i.mol for i in ins], 0)
    )
    if hasattr(self, '_dmol'):
        old = self._dmol
        f = bst.PhasePartition.dmol_relaxation_factor
        self._dmol = dmol = f * old + (1 - f) * new
    else:
        self._dmol = dmol = new
    self._dmol[np.abs(dmol) < 1e-9] = 0.
    self._duty = self.Hnet
    Ts_new = [i.T for i in outs]
    if all([i == j for i, j in zip(Ts_new, Ts)]): # T constant
        self._energy_variable = None
    elif self.Hnet == Q: # Q constant    
        self._energy_variable = 'T'
    else:
        self._energy_variable = None

class Unit(AbstractUnit):
    """
    Abstract class for Unit objects. Child objects must contain
    :attr:`~Unit._run`, :attr:`~Unit._design` and :attr:`~Unit._cost` methods to 
    estimate stream outlets of a Unit and find design and cost information.  

    Parameters
    ----------
    ID :
        A unique identification. If ID is None, unit will not be
        registered in flowsheet. By default, a unique ID will be chosen.
    ins :
        Inlet streams or IDs to initialize inlet streams.
        If empty tuple, streams with default IDs will be created.
        By default, streams will be missing.
    outs : 
        Outlet streams or IDs to initialize outlet streams.
        By default, streams with unique IDs will be created.
        If None, streams will be missing.
    thermo : 
        Thermo object to initialize inlet and outlet streams. Defaults to
        :meth:`settings.thermo <thermosteam._settings.ProcessSettings.thermo>`.
    
    Notes
    -----
    The free on board (f.o.b.) purchase costs and installed equipment costs 
    (i.e. bare-module cost) for each item in the :attr:`~Unit.baseline_purchase_costs` 
    dictionary and in auxiliary units are automatically added to the 
    :attr:`~Unit.purchase_costs` and :attr:`~Unit.installed_costs` dictionaries. 
    
    As explained in [1]_, the f.o.b. purchase cost is given by:
    
    .. math::
       
       C_{P} = C_{Pb}F_{D}F_{P}F_{M}
    
    And the installed equipment cost is given by:
    
    .. math::
       
       C_{BM} = C_{Pb} (F_{BM} + F_{D}F_{P}F_{M} - 1)
    
    Where:
        * :math:`C_{Pb}`: Baseline purchase cost.
        * :math:`F_{BM}`: Bare module factor.
        * :math:`F_{D}`: Design factor.
        * :math:`F_{P}`: Pressure factor.
        * :math:`F_{M}`: Material factor.
    
    Values for the bare-module, design, pressure, and material factors of 
    each equipment should be stored in the :attr:`~Unit.F_BM`, :attr:`~Unit.F_D`, 
    :attr:`~Unit.F_P`, and :attr:`~Unit.F_M` dictionaries.
    
    Examples
    --------
    :doc:`../tutorial/Creating_a_Unit`
    
    :doc:`../tutorial/-pipe-_notation`
    
    :doc:`../tutorial/Inheriting_from_Unit`
    
    References
    ----------
    .. [1] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani, R., & Ng, M. K. (2017). Product and Process Design Principles. Wiley. Cost Accounting and Capital Cost Estimation (Chapter 16)
    
    
    """ 
    Stream = Stream
    max_parallel_units = int(10e3)
    
    def __init_subclass__(cls,
                          isabstract=False,
                          new_graphics=True,
                          does_nothing=None,
                          default_phenomena=None):
        super().__init_subclass__()
        if does_nothing: return 
        dct = cls.__dict__
        if 'run' in dct:
            raise UnitInheritanceError(
                 "the 'run' method cannot be overridden; implement `_run` instead"
            )
        if default_phenomena is None:
            methods = (
                '_update_nonlinearities',
                '_get_energy_departure_coefficient',
                '_create_material_balance_equations',
                '_create_energy_departure_equations',
                '_update_energy_variable',
                '_energy_variable',
            )
            if all([getattr(cls, i) is getattr(Unit, i) for i in methods]):
                cls.run = phenomena_oriented_run
            else:
                for i in methods: 
                    if getattr(cls, i) is not getattr(Unit, i): continue
                    setattr(cls, i, AbstractMethod)
                cls.run = Unit.run
        elif default_phenomena:
            cls.run = phenomena_oriented_run
            for i in methods: setattr(cls, i, getattr(Unit, i))
        if '_N_heat_utilities' in dct:
            warn("'_N_heat_utilities' class attribute is scheduled for deprecation; "
                 "use the `add_heat_utility` method instead",
                 DeprecationWarning, stacklevel=2)
        if 'line' not in dct:
            cls.line = format_title(cls.__name__)
        if 'ticket_name' not in dct:
            line = cls.line.lower()
            if 'centrifuge' in line: cls.ticket_name = 'C'
            elif 'distillation' in line: cls.ticket_name = 'D'
            elif 'evaporator' in line: cls.ticket_name = 'E'
            elif 'flash' in line: cls.ticket_name = 'F'
            elif ('cooler' in line 
                  or 'condenser' in line
                  or 'heater' in line 
                  or 'boiler' in line
                  or 'heat exchanger' in line): cls.ticket_name = 'H'
            elif 'compressor' in line: cls.ticket_name = 'K'
            elif 'turbine' in line: cls.ticket_name = 'Ʞ'
            elif 'mixer' in line: cls.ticket_name = 'M'
            elif 'pump' in line: cls.ticket_name = 'P'
            elif 'reactor' in line or 'digestion' in line or 'ferment' in line: cls.ticket_name = 'R'
            elif 'split' in line: cls.ticket_name = 'S'
            elif 'tank' in line: cls.ticket_name = 'T'
            elif 'junction' == line: cls.ticket_name = 'J'
            elif 'specification' in line: cls.ticket_name = 'PS'
            elif 'valve' in line: cls.ticket_name = 'V'
            else: cls.ticket_name = 'U'
        if '_graphics' not in dct and new_graphics:
            # Set new graphics for specified line
            cls._graphics = UnitGraphics.box(cls._N_ins, cls._N_outs)
        if not isabstract:
            if hasattr(cls, '_BM'): 
                raise UnitInheritanceError(
                    'cannot set `_BM`; implement `_F_BM_default` instead'
                )
            elif hasattr(cls, '_F_BM_defaults'):
                raise UnitInheritanceError(
                    'cannot set `_F_BM_defaults`; implement '
                    '`_F_BM_default` instead'
                )
            elif cls._F_BM_default is Unit._F_BM_default:
                cls._F_BM_default = {}
            
            if hasattr(cls, '_equipment_lifetime'):
                raise UnitInheritanceError(
                    'cannot set `_equipment_lifetime`; '
                    'implement `_default_equipment_lifetime` instead'
                )
            elif hasattr(cls, '_default_equipment_lifetimes'):
                raise UnitInheritanceError(
                    'cannot set `_default_equipment_lifetimes`; implement '
                    '`_default_equipment_lifetime` instead'
                )
            elif cls._default_equipment_lifetime is Unit._default_equipment_lifetime: 
                cls._default_equipment_lifetime = {}
            if cls._units is Unit._units: cls._units = {}
            if not cls._run:
                if cls._N_ins == 1 and cls._N_outs == 1:
                    static(cls)
                else:
                    raise UnitInheritanceError(
                        "Unit subclass with multiple inlet or outlet streams "
                        "must implement a '_run' method unless the "
                        "'isabstract' keyword argument is True"
                    )
        name = cls.__name__
        if hasattr(bst, 'units') and hasattr(bst, 'wastewater') and hasattr(bst, 'facilities'):
            # Add 3rd party unit to biosteam module for convinience
            if name not in bst.units.__dict__:
                bst.units.__dict__[name] = cls
            if name not in bst.__dict__:
                bst.__dict__[name] = cls
        
    ### Abstract Attributes ###
    #: **class-attribute** Units of measure for :attr:`~Unit.design_results` dictionary.
    _units: dict[str, str] = {}
    
    #: **class-attribute** Default bare-module factors for each purchase cost item.
    #: Items in this dictionary are copied to the :attr:`~Unit.F_BM` attribute during 
    #: initialization.
    _F_BM_default: dict[str, float] = {}
    
    #: **class-attribute** Cost items that need to be summed across operation modes for 
    #: flexible operation (e.g., filtration membranes).
    _materials_and_maintenance: frozenset[str] = frozenset()

    #: **class-attribute** Whether to link inlet and outlet streams.
    _link_streams: bool = False

    #: **class-attribute** Lifetime of equipment. Defaults to lifetime of
    #: production venture. Use an integer to specify the lifetime for all
    #: items in the unit purchase costs. Use a dictionary to specify the 
    #: lifetime of each purchase cost item.
    _default_equipment_lifetime: int|dict[str, int] = {}

    #: [str] The energy variable for phenomena-oriented simulation.
    _energy_variable: str = None

    ### Abstract methods ###
    
    #: Create auxiliary components.
    _load_components = AbstractMethod
    
    #: Add design requirements to the :attr:`~Unit.design_results` dictionary.
    _design = AbstractMethod
    
    #: Add itemized purchase costs to the :attr:`~Unit.baseline_purchase_costs` dictionary.
    _cost = AbstractMethod

    #: Add embodied emissions (e.g., unit construction) in LCA
    _lca = AbstractMethod
    
    def initialize_equation_nodes(self):
        for eq in self.equation_node_names: getattr(self, f'initialize_{eq}')()
    
    def create_equation_nodes(self):
        eqs = self.equation_node_names
        equation_nodes = []
        for eq in eqs: 
            node = EquationNode(f"{self.node_tag}.{eq}")
            equation_nodes.append(node)
            setattr(self, eq, node)
        self.equation_nodes = tuple(equation_nodes)
        if not hasattr(self, 'energy_balance_node'):
            self.energy_balance_node = None
    
    @property
    def node_tag(self):
        if self.owner is self:
            return self.ID
        else:
            return f"{self.owner.node_tag}.{self.ID}"
        
    def _update_nonlinearities(self):
        """
        Update phenomenological variables for phenomena-oriented simulation.
        """
        ins = self.ins
        outs = self.outs
        data = [i.get_data() for i in outs]
        Ts = [i.T for i in outs]
        Q = self.Hnet
        self._run()
        f = bst.PhasePartition.dmol_relaxation_factor
        old = self._dmol
        new = sum(
            [i.mol for i in outs],
            -sum([i.mol for i in ins], 0)
        )
        self._dmol = dmol = f * old + (1 - f) * new
        self._dmol[np.abs(dmol) < 1e-9] = 0.
        self._duty = self.Hnet
        Ts_new = [i.T for i in outs]
        if all([i == j for i, j in zip(Ts_new, Ts)]): # T constant
            self._energy_variable = None
        elif self.Hnet == Q: # Q constant    
            self._energy_variable = 'T'
        else:
            self._energy_variable = None
        for i, j in zip(outs, data): i.set_data(j)
    
    def _simulation_error(self):
        ins = self.ins
        outs = self.outs
        new_ins = [i.copy() for i in ins]
        new_outs = [i.copy() for i in outs]
        streams = new_ins + new_outs
        Ts = np.array([i.T for i in streams])
        flows = np.array([i.mol for i in streams])
        self._ins = new_ins
        self._outs = new_outs
        self._run()
        Ts_new = np.array([i.T for i in streams])
        new_flows = np.array([i.mol for i in streams])
        self._ins = ins
        self._outs = outs
        return np.abs(new_flows - flows).sum(), np.abs(Ts_new - Ts).sum()
    
    def _get_energy_departure_coefficient(self, stream):
        """
        tuple[object, float] Return energy departure coefficient of a stream 
        for phenomena-oriented simulation.
        """
        if self._energy_variable == 'T': return (self, stream.C)
    
    def _create_material_balance_equations(self, composition_sensitive):
        """
        list[tuple[dict, array]] Create material balance equations for 
        phenomena-oriented simulation.
        """
        if self._link_streams: return []
        fresh_inlets, process_inlets, equations = self._begin_equations(composition_sensitive)
        outs = self.flat_outs
        N = self.chemicals.size
        ones = np.ones(N)
        predetermined_flow = SparseVector.from_dict(sum_sparse_vectors([i.mol for i in fresh_inlets]), size=N)
        rhs = predetermined_flow + self._dmol
        mol_total = sum([i.mol for i in outs])
        for i, s in enumerate(outs):
            split = s.mol / mol_total
            minus_split = -split
            eq_outs = {}
            for i in process_inlets: eq_outs[i] = minus_split
            eq_outs[s] = ones
            equations.append(
                (eq_outs, split * rhs)
            )
        return equations
    
    def _create_energy_departure_equations(self):
        """
        list[tuple[dict, float]] Create energy departure equations for 
        phenomena-oriented simulation.
        """
        if self._energy_variable == 'T':
            coeff = {self: sum([i.C for i in self.outs])}
            for i in self.ins: i._update_energy_departure_coefficient(coeff)
            if self._dmol.any():
                dH = self._duty - self.Hnet
            else:
                dH = self._duty + self.H_in - self.H_out
            return [(coeff, dH)]
        else:
            return []
    
    def _update_energy_variable(self, departure):
        """
        Update energy variable being solved in energy departure equations for 
        phenomena-oriented simulation.
        """
        if self._energy_variable == 'T':
            for i in self.outs: i.T += departure
    
    def _begin_equations(self, composition_sensitive):
        inlets = self.ins
        fresh_inlets = []
        process_inlets = []
        material_equations = [f() for i in inlets for f in i.material_equations]
        isfeed = lambda s: s.isfeed() or i.source._recycle_system is not self._recycle_system
        if not material_equations:
            if composition_sensitive:
                for i in inlets: 
                    if (isfeed(i) or
                        not getattr(i.source, 'composition_sensitive', False)):
                        fresh_inlets.append(i)
                    elif len(i) > 1:
                        process_inlets.extend(i)
                    else:
                        process_inlets.append(i)                    
            else:
                for i in inlets: 
                    if isfeed(i):
                        fresh_inlets.append(i)
                    elif len(i) > 1:
                        process_inlets.extend(i)
                    else:
                        process_inlets.append(i)   
            equations = material_equations
        elif composition_sensitive:
            equations = []
            dependent_streams = []
            for eq in material_equations:
                dct, value = eq 
                for i in dct:
                    if (i.source and
                        i.source._recycle_system is self._recycle_system and 
                        not getattr(i.source, 'composition_sensitive', False)):
                        break
                else:
                    dependent_streams.extend([i.imol for i in dct])
                    equations.append(eq)
            dependent_streams = set(dependent_streams)
            for i in inlets: 
                if (i.imol not in dependent_streams and isfeed(i) or
                    not getattr(i.source, 'composition_sensitive', False)):
                    fresh_inlets.append(i)
                elif len(i) > 1:
                    process_inlets.extend(i)
                else:
                    process_inlets.append(i)   
        else:
            for i in inlets: 
                if (isfeed(i) and 
                    not i.material_equations):
                    fresh_inlets.append(i)
                elif len(i) > 1:
                    process_inlets.extend(i)
                else:
                    process_inlets.append(i)   
            equations = material_equations
        return fresh_inlets, process_inlets, equations
    
    Inlets = piping.Inlets
    Outlets = piping.Outlets
    
    def __init__(self,
            ID: Optional[str]='',
            ins: streams=None,
            outs: streams=(),
            thermo: Optional[tmo.Thermo]=None,
            **kwargs
        ):
        
        ### Initialize utilities
    
        #: All heat utilities associated to unit. Cooling and heating requirements 
        #: are stored here (including auxiliary requirements).
        self.heat_utilities: list[HeatUtility, ...] = [HeatUtility for i in range(getattr(self, '_N_heat_utilities', 0))]
        
        #: Electric utility associated to unit (including auxiliary requirements).
        self.power_utility: PowerUtility = PowerUtility()
    
        ### Initialize design/cost/LCA results
        
        try:
            #: All bare-module factors for each purchase cost. Defaults to values in 
            #: the class attribute :attr:`~Unit._F_BM_default`.
            self.F_BM: dict[str, float] = self._F_BM_default.copy()
        except AttributeError:
            self.F_BM = {}
        
        #: All design factors for each purchase cost item in :attr:`~Unit.baseline_purchase_costs`.
        self.F_D: dict[str, float] = {}
        
        #: All pressure factors for each purchase cost item in :attr:`~Unit.baseline_purchase_costs`.
        self.F_P: dict[str, float] = {}
        
        #: All material factors for each purchase cost item in :attr:`~Unit.baseline_purchase_costs`.
        self.F_M: dict[str, float] = {}
        
        #: All design requirements excluding utility requirements and detailed 
        #: auxiliary unit requirements.
        self.design_results: dict[str, object] = {}
        
        #: All baseline purchase costs without accounting for design, 
        #: pressure, and material factors.
        self.baseline_purchase_costs: dict[str, float] = {}
        
        #: Itemized purchase costs (including auxiliary units)
        #: accounting for design, pressure, and material factors (i.e., 
        #: :attr:`~Unit.F_D`, :attr:`~Unit.F_P`, :attr:`~Unit.F_M`).
        #: Items here are automatically updated at the end of unit simulation.
        self.purchase_costs: dict[str, float] = {}
        
        #: All installed costs accounting for bare module, design, 
        #: pressure, and material factors. Items here are automatically updated
        #: at the end of unit simulation.
        self.installed_costs: dict[str, float] = {}
        
        #: Indices of additional utilities given by inlet streams.
        self._inlet_utility_indices: dict[str, int] = {}
        
        #: Indices of additional utilities given by outlet streams.
        self._outlet_utility_indices: dict[str, int] = {}
        
        #: Indices of additional credits/fees given by inlet streams.
        self._inlet_cost_indices: dict[str, int] = {}
        
        #: Indices of additional credits/fees given by outlet streams.
        self._outlet_revenue_indices: dict[str, int] = {}
        
        try:
            #: Lifetime of equipment. Defaults to values in the class attribute 
            #: :attr:`~Unit._default_equipment_lifetime`. Use an integer to specify the lifetime 
            #: for all items in the unit purchase costs. Use a dictionary to specify 
            #: the lifetime of each purchase cost item.
            self.equipment_lifetime: int|dict[str, int] = copy(self._default_equipment_lifetime)
        except AttributeError:
            self.equipment_lifetime = {}
        
        #: Name-number pairs of baseline purchase costs and auxiliary unit 
        #: operations in parallel. Use 'self' to refer to the main unit. Capital 
        #: and heat and power utilities in parallel will become proportional to this 
        #: value.
        self.parallel: dict[str, int] = {}
        
        #: Unit design decisions that must be solved to satisfy specifications.
        #: While adding responses is optional, simulations benefit from responses
        #: by being able to predict better guesses.
        self.responses: set[bst.GenericResponse] = set()
        
        self._utility_cost = None
        
        self._recycle_system = None
        
        super().__init__(ID, ins, outs, thermo, **kwargs)
    
        self._assert_compatible_property_package()
        
    def _init_ins(self, ins):
        self.auxins = {}
        self._init_inlets(ins)
    
    def _init_outs(self, outs):
        self.auxouts = {}
        self._init_outlets(outs)

    def _init_utils(self):
        self.heat_utilities = [HeatUtility for i in range(getattr(self, '_N_heat_utilities', 0))]
        self.power_utility = PowerUtility()
        
    def _init_results(self):
        try: self.F_BM = self._F_BM_default.copy()
        except AttributeError: self.F_BM = {}
        self.F_D = {}
        self.F_P = {}
        self.F_M = {}
        self.design_results = {}
        self.baseline_purchase_costs = {}
        self.purchase_costs = {}
        self.installed_costs = {}
        self._inlet_utility_indices = {}
        self._outlet_utility_indices = {}
        self._inlet_cost_indices = {}
        self._outlet_revenue_indices = {}
        try: self.equipment_lifetime = copy(self._default_equipment_lifetime)
        except AttributeError: self.equipment_lifetime = {}

    def _init_specifications(self):
        self._specifications = []
        self.run_after_specifications = False
        self._active_specifications = set()
    
    def _reset_thermo(self, thermo):
        for i in (self._ins._streams + self._outs._streams):
            try:
                if i: i._reset_thermo(thermo)
            except:
                raise RuntimeError(f'failed to reset {repr(self)}.thermo')
        if thermo is self.thermo: return
        self._load_thermo(thermo)
        chemicals = thermo.chemicals
        isa = isinstance
        hasfield = hasattr
        def reset_thermo(obj, old=set()):
            hash = id(obj)
            if hash in old: return 
            old.add(hash)
            if isa(obj, tmo.ReactionSystem):
                for rxn in obj._reactions:
                    if hasfield(rxn, 'reset_chemicals') and rxn.chemicals is not chemicals:
                        rxn.reset_chemicals(chemicals)
            elif hasfield(obj, 'reset_chemicals') and obj.chemicals is not chemicals:
                obj.reset_chemicals(chemicals)
            elif hasfield(obj, '_reset_thermo') and obj.thermo is not thermo:
                obj._reset_thermo(thermo)
            elif isa(obj, dict):
                for i in obj.values(): reset_thermo(i)
            elif isa(obj, Iterable):
                for i in obj: reset_thermo(i)
                
        for obj in self.__dict__.values(): reset_thermo(obj)
            
    def response(self, name):
        """Register response for convergence model prediction."""
        self.responses.add(
            bst.GenericResponse(self, name)
        )
            
    @property
    def net_power(self) -> float:
        """Net power consumption [kW]."""
        return self.power_utility.power
    @property
    def net_duty(self) -> float:
        """Net duty including heat transfer losses [kJ/hr]."""
        return sum([i.duty for i in self.heat_utilities])
    @property
    def net_cooling_duty(self) -> float:
        """Net cooling duty including heat transfer losses [kJ/hr]."""
        return sum([i.duty for i in self.heat_utilities if i.duty < 0.])
    @property
    def net_heating_duty(self) -> float:
        """Net cooling duty including heat transfer losses [kJ/hr]."""
        return sum([i.duty for i in self.heat_utilities if i.duty > 0.])
    
    def add_power_utility(self, power):
        """Add power utility [kW]. Use a positive value for consumption and 
        a negative for production."""
        power_utility = self.power_utility
        if power >= 0.:
            power_utility.consumption += power
        else:
            power_utility.production -= power
    
    def create_heat_utility(self,
            agent: Optional[UtilityAgent]=None,
            heat_transfer_efficiency: Optional[float]=None,
        ):
        """Create heat utility object associated to unit."""
        hu = HeatUtility(heat_transfer_efficiency, None)
        self.heat_utilities.append(hu)
        return hu
    
    def add_heat_utility(self, 
            unit_duty: float, 
            T_in: float,
            T_out: Optional[float]=None, 
            agent: Optional[UtilityAgent]=None,
            heat_transfer_efficiency: Optional[float]=None,
            hxn_ok: Optional[bool]=False,
        ):
        """
        Add utility requirement given the duty and inlet and outlet 
        temperatures.
        
        Parameters
        ----------
        unit_duty :
               Unit duty requirement [kJ/hr]
        T_in : 
               Inlet process stream temperature [K]
        T_out : 
               Outlet process stream temperature [K]
        agent : 
                Utility agent to use. Defaults to a suitable agent from 
                predefined heating/cooling utility agents.
        heat_transfer_efficiency : 
            Enforced fraction of heat transfered from utility (due
            to losses to environment).
        hxn_ok :
            Whether heat utility can be satisfied within a heat exchanger network.
            
        """
        hu = HeatUtility(heat_transfer_efficiency, self, hxn_ok)
        self.heat_utilities.append(hu)
        hu(unit_duty, T_in, T_out, agent)
        return hu
    
    def define_utility(self, name: str, stream: Stream):
        """
        Define an inlet or outlet stream as a utility by name.
        
        Parameters
        ----------
        name : 
            Name of utility, as defined in :meth:`settings.stream_prices <thermosteam._settings.ProcessSettings.stream_prices>`.
        stream :
            Inlet or outlet utility stream.
        
        """
        if name not in bst.stream_prices:
            raise ValueError(f"price of '{name}' must be defined in settings.stream_prices")
        if stream._sink is self:
            self._inlet_utility_indices[name] = self._ins._streams.index(stream)
        elif stream._source is self:
            self._outlet_utility_indices[name] = self._outs._streams.index(stream)
        else:
            raise ValueError(f"stream '{stream.ID}' must be connected to {repr(self)}")
    
    def define_credit(self, name: str, stream: Stream):
        """
        Define an inlet or outlet stream as a fee/credit by name.
        
        Parameters
        ----------
        name : 
            Name of fee/credit, as defined in :meth:`settings.stream_prices <thermosteam._settings.ProcessSettings.stream_prices>`.
        stream :
            Inlet or outlet fee/credit stream.
        
        """
        if name not in bst.stream_prices:
            raise ValueError(f"price of '{name}' must be defined in settings.stream_prices")
        if stream._sink is self:
            self._inlet_cost_indices[name] = self._ins._streams.index(stream)
        elif stream._source is self:
            self._outlet_revenue_indices[name] = self._outs._streams.index(stream)
        else:
            raise ValueError(f"stream '{stream.ID}' must be connected to {repr(self)}")
    
    define_fee = define_credit
    
    def get_inlet_cost_flows(self):
        ins = self._ins._streams
        return {name: ins[index].F_mass for name, index in (self._inlet_utility_indices | self._inlet_cost_indices).items()}
    
    def get_outlet_revenue_flows(self):
        outs = self._outs._streams
        return {name: outs[index].F_mass for name, index in (self._outlet_utility_indices | self._outlet_revenue_indices).items()}
    
    def get_design_and_capital(self):
        return UnitDesignAndCapital(
            self, self.F_BM.copy(), self.F_D.copy(), self.F_P.copy(), self.F_M.copy(), 
            self.design_results.copy(), self.baseline_purchase_costs.copy(),
            self.purchase_costs.copy(), self.installed_costs.copy(),
        )
    
    def get_agile_design_and_capital(self, design_and_capital: list[UnitDesignAndCapital]):
        names = (
            'F_BM', 'F_D', 'F_P', 'F_M', 'design_results',
            'baseline_purchase_costs', 'purchase_costs', 'installed_costs',
        )
        max_agile_design = getattr(self, '_max_agile_design', None)
        agile_scenario = {i: {} for i in names}
        if max_agile_design:
            designs = [i.design_results for i in design_and_capital]
            agile_design = agile_scenario['design_results']
            for design_results in designs:
                for i in max_agile_design:
                    if i in design_results:
                        j = design_results[i]
                    else:
                        continue
                    if i in agile_design:
                        if abs(j) > abs(agile_design[i]):
                            agile_design[i] = j
                    else:
                        agile_design[i] = j
            names = ('design_results', 'baseline_purchase_costs', 
                     'purchase_costs', 'installed_costs')
            dcts = [getattr(self, i) for i in names]
            self.design_results = agile_design
            for i in names[1:]: setattr(self, i, {}) 
            Unit._setup(self)
            try:
                self._cost()
                self._load_costs()
            except:
                warn(f"failed to create agile design for {repr(self)}; "
                      "assuming design with highest capital cost will do",
                      category=RuntimeWarning, stacklevel=2)
            else:
                design_and_capital.append(self.get_design_and_capital())
            finally:
                for i, j in zip(names, dcts): setattr(self, i, j) 
        self._fill_agile_design_and_capital(agile_scenario, design_and_capital)
        agile_scenario['unit'] = self
        return UnitDesignAndCapital(**agile_scenario)
                
    def _fill_agile_design_and_capital(self, agile_scenario, design_and_capital):
        for results in design_and_capital:
            for name, maxdct in agile_scenario.items():
                if name == 'design_results': continue
                dct = getattr(results, name)
                for i, j in dct.items():
                    if i in maxdct:
                        if abs(j) > abs(maxdct[i]):
                            maxdct[i] = j
                    else:
                        maxdct[i] = j
        F_BM = agile_scenario['F_BM']
        F_D = agile_scenario['F_D']
        F_P = agile_scenario['F_P']
        F_M = agile_scenario['F_M']
        baseline_purchase_costs = agile_scenario['baseline_purchase_costs']
        purchase_costs = agile_scenario['purchase_costs']
        installed_costs = agile_scenario['installed_costs']
        materials_and_maintenance = self._materials_and_maintenance
        for name, Cp in baseline_purchase_costs.items():
            F = F_D.get(name, 1.) * F_P.get(name, 1.) * F_M.get(name, 1.)
            installed_cost = Cp * (F_BM.get(name, 1.) + F - 1.)
            purchase_cost = Cp * F
            if installed_cost > installed_costs[name]:
                if name in materials_and_maintenance:
                    installed_costs[name] += installed_cost
                else:
                    installed_costs[name] = installed_cost
            if purchase_cost > purchase_costs[name]:
                if name in materials_and_maintenance:
                    purchase_costs[name] += purchase_cost
                else:
                    purchase_costs[name] = purchase_cost
    
    def _assert_compatible_property_package(self):
        if self.owner is not self: return
        CASs = self.chemicals.CASs
        streams = self._ins + self._outs
        assert all([s.chemicals.CASs == CASs for s in streams if s]), (
            "unit operation chemicals are incompatible with inlet and outlet streams; "
            "try using the `thermo` keyword argument to initialize the unit operation "
            "with a compatible thermodynamic property package"
        )
    
    def _load_costs(self):
        r"""
        Calculate and save free on board (f.o.b.) purchase costs and
        installed equipment costs (i.e. bare-module cost) for each item in the 
        :attr:`~Unit.baseline_purchase_costs` dictionary and in auxiliary units. This 
        method is run after the :attr:`~Unit._cost` method at the end of unit simulation.
        
        Notes
        -----
        As explained in [1]_, the f.o.b. purchase cost is given by:
        
        .. math::
           
           C_{P} = C_{Pb}F_{D}F_{P}F_{M}
        
        And the installed equipment cost is given by:
        
        .. math::
           
           C_{BM} = C_{Pb} (F_{BM} + F_{D}F_{P}F_{M} - 1)
        
        Where:
            * :math:`C_{Pb}`: Baseline purchase cost.
            * :math:`F_{BM}`: Bare module factor.
            * :math:`F_{D}`: Design factor.
            * :math:`F_{P}`: Pressure factor.
            * :math:`F_{M}`: Material factor.
        
        Values for the bare-module, design, pressure, and material factors of 
        each equipment should be stored in the :attr:`~Unit.F_BM`, :attr:`~Unit.F_D`, 
        :attr:`~Unit.F_P`, and :attr:`~Unit.F_M` dictionaries.
        
        Warning
        -------
        If an item is listed in the :attr:`~Unit.purchase_costs` dictionary but not in the
        :attr:`~Unit.baseline_purchase_costs` dictionary, the baseline purchase cost is 
        assumed to be the same as the purchase cost.
        
        References
        ----------
        .. [1] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani, R., & Ng, M. K. (2017). Product and Process Design Principles. Wiley. Cost Accounting and Capital Cost Estimation (Chapter 16)
        
        """
        if getattr(self, '_costs_loaded', False): return
        F_BM = self.F_BM
        F_D = self.F_D
        F_P = self.F_P
        F_M = self.F_M
        baseline_purchase_costs = self.baseline_purchase_costs
        purchase_costs = self.purchase_costs
        installed_costs = self.installed_costs
        parallel = self.parallel
        heat_utilities = self.heat_utilities
        power_utility = self.power_utility
        integer = int
        N_default = integer(parallel.get('self', 1))
        
        # Load main costs
        if N_default != 1:
            heat_utilities.extend((N_default - 1) * heat_utilities)
            power_utility.scale(N_default)
        for i in purchase_costs:
            if i not in baseline_purchase_costs:
                warning = RuntimeWarning(
                    f"Unit {self.ID}, adding items to the `purchase_costs` dictionary is "
                    "deprecated; add items to `baseline_purchase_costs` "
                    "dictionary instead"
                 )
                warn(warning)
                baseline_purchase_costs[i] = purchase_costs[i]
        for name, Cpb in baseline_purchase_costs.items(): 
            N = integer(parallel.get(name, N_default))
            if N == 1:
                if name in installed_costs and name in purchase_costs:
                    continue # Assume costs already added elsewhere using another method
            else:
                Cpb *= N
                baseline_purchase_costs[name] = Cpb
                if name in installed_costs and name in purchase_costs:
                    purchase_costs[name] *= N
                    installed_costs[name] *= N
                    continue # Assume costs already added elsewhere using another method
            F = F_D.get(name, 1.) * F_P.get(name, 1.) * F_M.get(name, 1.)
            try:
                installed_costs[name] = Cpb * (F_BM[name] + F - 1.)
            except KeyError:
                warn(f"the purchase cost item, '{name}', has "
                      "no defined bare-module factor in the "
                     f"'{type(self).__name__}.F_BM' dictionary; "
                      "bare-module factor now has a default value of 1",
                      RuntimeWarning)
                F_BM[name] = 1.
                installed_costs[name] = purchase_costs[name] = Cpb * F
            else:
                purchase_costs[name] = Cpb * F
        
        # Load auxiliary costs
        isa = isinstance
        for name, unit in self.get_auxiliary_units_with_names():
            if isa(unit, Unit):
                if not (unit._design or unit._cost): continue
            unit._load_costs() # Just in case user did not simulate or run summary.
            pname, *_ = name.split('[')
            N = integer(parallel.get(pname, N_default))
            if N == 1:
                heat_utilities.extend(unit.heat_utilities)
                power_utility.consumption += unit.power_utility.consumption
                power_utility.production += unit.power_utility.production
            elif N > self.max_parallel_units:
                raise RuntimeError(f'cannot have over a {self.max_parallel_units} unit operations in parallel')
            else:
                heat_utilities.extend(N * unit.heat_utilities)
                power_utility.consumption += N * unit.power_utility.consumption
                power_utility.production += N * unit.power_utility.production
            bpc_auxiliary = unit.baseline_purchase_costs
            pc_auxiliary = unit.purchase_costs
            ic_auxiliary = unit.installed_costs
            for i in bpc_auxiliary:
                j = ' - '.join([name.capitalize().replace('_', ' '), i])
                if j in baseline_purchase_costs: 
                    raise RuntimeError(
                        f"'{j}' already in `baseline_purchase_cost` "
                        f"dictionary of {repr(self)}; try using a different key"
                    )
                elif N == 1:
                    baseline_purchase_costs[j] = bpc_auxiliary[i]
                    purchase_costs[j] = pc_auxiliary[i]
                    installed_costs[j] = ic_auxiliary[i]
                else:
                    baseline_purchase_costs[j] = N * bpc_auxiliary[i]
                    purchase_costs[j] = N * pc_auxiliary[i]
                    installed_costs[j] = N * ic_auxiliary[i]
            
        self._costs_loaded = True
    
    def _setup(self):
        """Clear all results, setup up stream conditions and constant data, 
        and update system configuration based on units impacted by process 
        specifications. This method is run at the start of unit simulation, 
        before running mass and energy balances."""
        self.materialize_connections()
        self.power_utility.empty()
        for i in self.heat_utilities: i.empty()
        if not hasattr(self, '_N_heat_utilities'): self.heat_utilities.clear()
        self.parallel.clear()
        self.baseline_purchase_costs.clear()
        self.purchase_costs.clear()
        self.installed_costs.clear()
        self._costs_loaded = False
        for i in self.auxiliary_units: i._setup()
    
    def _check_setup(self):
        if any([self.power_utility, 
                self.heat_utilities, 
                self.baseline_purchase_costs, 
                self.purchase_costs, 
                self.installed_costs]):
            raise UnitInheritanceError(
               f'`{type(self).__name__}._setup` method did not clear unit results; '
                'a potential solution is to run `super()._setup()` in the `_setup` '
                'method of the unit subclass'
            )
    
    def _check_run(self):
        if any([self.power_utility, 
                self.heat_utilities, 
                self.baseline_purchase_costs, 
                self.purchase_costs, 
                self.installed_costs]):
            warn(
               f'`{type(self).__name__}._run` method added unit results '
                '(e.g., purchase costs, heat and power utilities); unit results '
                'should only be added in `_design` or `_cost` methods',
                RuntimeWarning
            )
    
    def materialize_connections(self):
        for s in self._ins + self._outs: 
            if not s: s.materialize_connection()
    
    @property
    def owner(self) -> Unit:
        owner = getattr(self, '_owner', None)
        return self if owner is None else owner.owner
    @owner.setter
    def owner(self, owner):
        if owner is not self: self._owner = owner
    
    def _get_tooltip_string(self, format=None, full=None):
        """Return a string that can be used as a Tippy tooltip in HTML output"""
        if format is None: format = bst.preferences.graphviz_format
        if full is None: full = bst.preferences.tooltips_full_results
        if format not in ('html', 'svg'): return ''
        if format == 'html' and full:
            results = self.results(include_installed_cost=True)
            tooltip = (
                " " + # makes sure graphviz does not try to parse the string as HTML
                results.to_html(justify='unset'). # unset makes sure that table header style can be overwritten in CSS
                replace("\n", "").replace("  ", "") # makes sure tippy.js does not add any whitespaces
            )
        else:
            newline = '<br>' if format == 'html' else '\n'
            electricity_consumption = self.power_utility.consumption
            electricity_production = self.power_utility.production
            cooling = self.net_cooling_duty / 1e3
            heating = self.net_heating_duty / 1e3
            utility_cost = self.utility_cost
            purchase_cost = int(float(self.purchase_cost))
            installed_cost = int(float(self.installed_cost))
            tooltip = ''
            if electricity_consumption:
                tooltip += f"{newline}Electricity consumption: {electricity_consumption:.3g} kW"
            if electricity_production:
                tooltip += f"{newline}Electricity production: {electricity_production:.3g} kW"
            if cooling:
                tooltip += f"{newline}Cooling duty: {cooling:.3g} MJ/hr"
            if heating:
                tooltip += f"{newline}Heating duty: {heating:.3g} MJ/hr"
            if utility_cost:
                tooltip += f"{newline}Utility cost: {utility_cost:.3g} USD/hr"
            if purchase_cost:
                tooltip += f"{newline}Purchase cost: {purchase_cost:,} USD"
            if installed_cost:
                tooltip += f"{newline}Installed equipment cost: {installed_cost:,} USD"
            if not tooltip: tooltip = 'No capital costs or utilities'
            elif tooltip: tooltip = tooltip.lstrip(newline)
            if format == 'html': tooltip = ' ' + tooltip
        return tooltip
    
    def get_design_result(self, key: str, units: str):
        """
        Return design result in a new set of units of measure.
            
        Parameters
        ----------
        key :
            Name of design result.
        units :
            Units of measure.
        
        Examples
        --------
        >>> import biosteam as bst
        >>> bst.settings.set_thermo(['Water'], cache=True)
        >>> feed = bst.Stream(None, Water=100)
        >>> tank = bst.StorageTank(None, feed)
        >>> tank.simulate()
        >>> tank.get_design_result('Total volume', 'm3')
        1214.19
        >>> tank.get_design_result('Total volume', 'L')
        1214191.0
        
        """
        return convert(self.design_results[key], self._units[key], units)
    
    
    def set_design_result(self, key: str, units: str, value: float):
        """
        Set design result in given the units of measure.
            
        Parameters
        ----------
        key :
            Name of design result.
        units :
            Units of measure.
        value:
            Value of the design result.
        
        Examples
        --------
        >>> import biosteam as bst
        >>> bst.settings.set_thermo(['Water'], cache=True)
        >>> feed = bst.Stream(None, Water=100)
        >>> tank = bst.StorageTank(None, feed)
        >>> tank.simulate()
        >>> tank.set_design_result('Total volume', 'm3', 1000)
        1000
        >>> tank.get_design_result('Total volume', 'm3')
        1000.0
        
        """
        self.design_results[key] = value = convert(value, units, self._units[key])
        return value
    
    def convert_design_result(self, key, units, value):
        """
        Convert design result in given units to the stored units of measure.
            
        Parameters
        ----------
        key :
            Name of design result.
        units :
            Units of measure.
        value:
            Value of the design result.
        
        Examples
        --------
        >>> import biosteam as bst
        >>> bst.settings.set_thermo(['Water'], cache=True)
        >>> feed = bst.Stream(None, Water=100)
        >>> tank = bst.StorageTank(None, feed)
        >>> tank.simulate()
        >>> tank.convert_design_result('Total volume', 'ft3', 1000)
        28.31
        
        """
        return convert(value, units, self._units[key])
    
    def reset_cache(self, isdynamic=None):
        pass
    
    def _get_design_info(self): 
        return ()
    
    def _load_stream_links(self):
        if self._link_streams: self._outs[0].link_with(self._ins[0])
    
    def _reevaluate(self):
        """Reevaluate design/cost/LCA results."""
        self._setup()
        self._summary()
    
    def _check_utilities(self):
        auxiliary_heat_utilities = set(sum([i.heat_utilities for i in self.auxiliary_units], []))
        for i in self.heat_utilities:
            if i in auxiliary_heat_utilities:
                raise UnitInheritanceError(
                    'auxiliary heat utilities were manually added to main utilities; '
                    'note that utilities from auxiliary units are already automatically '
                    'added to main unit operation'
                )
    
    def _summary(self, design_kwargs=None, cost_kwargs=None, lca_kwargs=None):
        """Run design/cost/LCA algorithms and compile results."""
        self._check_run()
        if not (self._design or self._cost): return
        if not self._skip_simulation_when_inlets_are_empty or not all([i.isempty() for i in self._ins]): 
            self._design(**design_kwargs) if design_kwargs else self._design()
            self._cost(**cost_kwargs) if cost_kwargs else self._cost()
            self._lca(**lca_kwargs) if lca_kwargs else self._lca()
            self._check_utilities()
        self._load_costs()
        self._load_operation_costs()

    def _load_operation_costs(self):
        ins = self._ins._streams
        outs = self._outs._streams
        prices = bst.stream_prices
        self._utility_cost = (
            sum([i.cost for i in self.heat_utilities]) 
            + self.power_utility.cost
            + sum([s.F_mass * prices[name] for name, index in self._inlet_utility_indices.items() if (s:=ins[index]).price == 0.])
            - sum([s.F_mass * prices[name] for name, index in self._outlet_utility_indices.items() if (s:=outs[index]).price == 0.])
        )
        self._inlet_cost = sum(
            [ins[index].F_mass * prices[name] for name, index in self._inlet_cost_indices.items()]
        )
        self._outlet_revenue = sum(
            [outs[index].F_mass * prices[name] for name, index in self._outlet_revenue_indices.items()]
        )
    
    @property
    def specifications(self) -> list[ProcessSpecification]:
        """
        Process specifications as a list of process specification objects.
        
        See Also
        --------
        add_specification
        add_bounded_numerical_specification
        
        """
        return self._specifications
    @specifications.setter
    def specifications(self, specifications):
        if specifications is None:
            self._specifications = []
        else:
            self._specifications = specifications
    
    @property
    def specification(self):
        raise AttributeError('`specification` property is deprecated; use `add_specification` or `specifications` (plural with an s) instead')
    @specification.setter
    def specification(self, specification):
        raise AttributeError('`specification` property is deprecated; use `add_specification` or `specifications` (plural with an s) instead')
    
    @property
    def baseline_purchase_cost(self) -> float:
        """Total baseline purchase cost, without accounting for design ,
        pressure, and material factors [USD]."""
        return sum(self.baseline_purchase_costs.values())
    
    @property
    def purchase_cost(self) -> float:
        """Total purchase cost [USD]."""
        return sum(self.purchase_costs.values())
    
    @property
    def installed_cost(self) -> float:
        """Total installed equipment cost [USD]."""
        return sum(self.installed_costs.values())
    
    @property
    def utility_cost(self) -> float:
        """Total utility cost [USD/hr]."""
        return self._utility_cost

    def mass_balance_error(self):
        """Return error in stoichiometric mass balance. If positive,
        mass is being created. If negative, mass is being destroyed."""
        return self.F_mass_out - self.F_mass_in

    def atomic_balance_error(self):
        """Return a dictionary of errors in stoichiometric atomic balances. 
        If value is positive, the atom is being created. If negative, the atom 
        is being destroyed."""
        from chemicals import elements
        mol = sum([i.mol for i in self._outs._streams]) - sum([i.mol for i in self._ins._streams])
        formula_array = self.chemicals.formula_array
        unbalanced_array = formula_array @ mol
        return elements.array_to_atoms(unbalanced_array)

    def empty(self):
        """
        Empty all unit operation results and outlet flows.
        """
        self.design_results.clear()
        self.baseline_purchase_costs.clear()
        self.purchase_costs.clear()
        self.installed_costs.clear()
        for i in self.outs: i.empty()
        self.heat_utilities.clear()
        self.power_utility.empty()
        self._utility_cost = 0.
        
    def simulate(self, 
            run: Optional[bool]=None,
            design_kwargs: Optional[dict]=None,
            cost_kwargs: Optional[dict]=None):
        """
        Run rigorous simulation and determine all design requirements.
        
        Parameters
        ----------
        run :
            Whether to run mass and energy balance or to assume the same inlet
            and outlet conditions. Defaults to True.
        design_kwargs :
            Keyword arguments passed to `_design` method.
        cost_kwargs :
            Keyword arguments passed to `_cost` method.
            
        """
        self._setup()
        self._check_setup()
        if run is None or run:
            for ps in self._specifications: ps.compile_path(self)
            self._load_stream_links()
            self.run()
        self._summary(design_kwargs, cost_kwargs)

    def _mass_and_energy_balance_specifications(self):
        return (self.line, ())

    def _mass_and_energy_balance_specifications_table(self):
        title, specs = self._mass_and_energy_balance_specifications()
        index = []
        values = []
        for name, value, units in specs:
            index.append(name)
            values.append([value, units])
        df = pd.DataFrame(
            values, index, ('Values', 'Units')
        )
        df.columns.name = f"{title} - {self.ID}"
        return df

    def results(self, with_units=True, include_utilities=True,
                include_total_cost=True, include_installed_cost=False,
                include_zeros=True, external_utilities=None, key_hook=None,
                basis=None):
        """
        Return key results from simulation as a DataFrame if `with_units`
        is True or as a Series otherwise.
        """
        def addkey(key):
            if key_hook: key = key_hook(key)
            keys.append(key)
            
        def addcapex(key):
            if key_hook: key = key_hook(key)
            *others, name = key
            names = name.split(' - ')
            parent = self
            parallel = parent.parallel
            N = N_default = parallel.get('self', 1)
            for auxname in names:
                auxsearch = auxname.replace(' ', '_').lower()
                N *= int(parallel.get(auxname, 1.))
                for i in parent.auxiliary_unit_names:
                    if auxsearch == i.lower():
                        parent = getattr(parent, i)
                        if isinstance(parent, list):
                            N *= len(parent)
                        else:
                            parallel = parent.parallel
                            N_default = parallel.get('self', 1)
                            N *= int(parallel.get(i, N_default))
                        break
                else:
                    break
            if N != 1: key = (*others, name + f' (x{N})')
            keys.append(key)
        keys = []; 
        vals = []; addval = vals.append
        stream_prices = bst.stream_prices
        all_utilities = self.heat_utilities + external_utilities if external_utilities else self.heat_utilities
        if with_units:
            if include_utilities:
                power_utility = self.power_utility
                if power_utility:
                    addkey(('Electricity', 'Power'))
                    addval(('kW', power_utility.power))
                    if include_zeros or power_utility.cost:
                        addkey(('Electricity', 'Cost'))
                        addval(('USD/hr', power_utility.cost))
                for heat_utility in HeatUtility.sum_by_agent(all_utilities):
                    if heat_utility:
                        ID = heat_utility.ID.replace('_', ' ').capitalize()
                        addkey((ID, 'Duty'))
                        addval(('kJ/hr', heat_utility.duty))
                        addkey((ID, 'Flow'))
                        addval(('kmol/hr', heat_utility.flow))
                        if include_zeros or heat_utility.cost: 
                            addkey((ID, 'Cost'))
                            addval(('USD/hr', heat_utility.cost))
                for name, flow in self.get_inlet_cost_flows().items():
                    if include_zeros or flow:
                        ID = name + ' (inlet)'
                        addkey((ID, 'Flow'))
                        addval(('kg/hr', flow))
                        addkey((ID, 'Cost'))
                        addval(('USD/hr', flow * stream_prices[name]))
                for name, flow in self.get_outlet_revenue_flows().items():
                    if include_zeros or flow:
                        ID = name + ' (outlet)'
                        addkey((ID, 'Flow'))
                        addval(('kg/hr', flow))
                        addkey((ID, 'Cost'))
                        addval(('USD/hr', - flow * stream_prices[name]))
            units = self._units
            Cost = self.purchase_costs
            if basis:
                ureg = tmo.units_of_measure.ureg
                ureg.default_system = basis
                Quantity = ureg.Quantity
                for ki, vi in self.design_results.items():
                    addkey(('Design', ki))
                    if ki in units:
                        ui = units[ki]
                        if ui != 'hr':
                            q = Quantity(vi, ui)
                            q = q.to_base_units()
                            vi = q.magnitude
                            ui = q.units
                    else:
                        ui = ''
                    addval((ui, vi))
                for ki, vi, ui in self._get_design_info():
                    addkey(('Design', ki))
                    q = Quantity(vi, ui)
                    q = q.to_base_units()
                    vi = q.magnitude
                    ui = q.units
                    addval((ui, vi))
            else:
                for ki, vi in self.design_results.items():
                    addkey(('Design', ki))
                    addval((units.get(ki, ''), vi))
                for ki, vi, ui in self._get_design_info():
                    addkey(('Design', ki))
                    addval((ui, vi))
            for ki, vi in Cost.items():
                addcapex(('Purchase cost', ki))
                addval(('USD', vi))
            if include_total_cost:
                addkey(('Total purchase cost', ''))
                addval(('USD', self.purchase_cost))
                if include_installed_cost:
                    addkey(('Installed equipment cost', ''))
                    addval(('USD', self.installed_cost))
                utility_cost = self.utility_cost
                if include_zeros or utility_cost: 
                    addkey(('Utility cost', ''))
                    addval(('USD/hr', utility_cost))
            if not keys: return None
            df = pd.DataFrame(vals,
                              pd.MultiIndex.from_tuples(keys),
                              ('Units', self.ID))
            df.columns.name = self.line
            return df
        else:
            if include_utilities:
                power_utility = self.power_utility
                if power_utility:
                    addkey(('Electricity', 'Power'))
                    addval(power_utility.power)
                    if include_zeros or power_utility.cost:
                        addkey(('Electricity', 'Cost'))
                        addval(power_utility.cost)
                for heat_utility in HeatUtility.sum_by_agent(all_utilities):
                    if heat_utility:
                        ID = heat_utility.ID.replace('_', ' ').capitalize()
                        addkey((ID, 'Duty'))
                        addval(heat_utility.duty)
                        addkey((ID, 'Flow'))
                        addval(heat_utility.flow)
                        if include_zeros or heat_utility.cost:
                            addkey((ID, 'Cost'))
                            addval(heat_utility.cost)
                for name, flow in self.get_inlet_cost_flows().items():
                    if include_zeros or flow:
                        ID = name + ' (inlet)'
                        addkey((ID, 'Flow'))
                        addval(flow)
                        addkey((ID, 'Cost'))
                        addval(flow * stream_prices[name])
                for name, flow in self.get_outlet_revenue_flows().items():
                    if include_zeros or flow:
                        ID = name + ' (outlet)'
                        addkey((ID, 'Flow'))
                        addval(flow)
                        addkey((ID, 'Cost'))
                        addval(-flow * stream_prices[name])
                            
            for ki, vi in self.design_results.items():
                addkey(('Design', ki))
                addval(vi)
            for ki, vi, ui in self._get_design_info():
                addkey(('Design', ki))
                addval(vi)
            for ki, vi in self.purchase_costs.items():
                addcapex(('Purchase cost', ki))
                addval(vi)
            if include_total_cost:
                addkey(('Total purchase cost', ''))
                addval(self.purchase_cost)
                if include_installed_cost:
                    addkey(('Installed equipment cost', ''))
                    addval(self.installed_cost)
                utility_cost = self.utility_cost
                if include_zeros or utility_cost:
                    addkey(('Utility cost', ''))
                    addval(utility_cost)
            if not keys: return None
            series = pd.Series(vals, pd.MultiIndex.from_tuples(keys))
            series.name = self.ID
            return series

    def get_available_chemicals(self):
        streams = [i for i in (self._ins + self._outs) if i]
        reaction_chemicals = sum([i.reaction_chemicals for i in self.__dict__.values() if hasattr(i, 'reaction_chemicals')], [])
        required_chemicals = set(sum([i.available_chemicals for i in streams], reaction_chemicals))
        return [i for i in self.chemicals if i in required_chemicals]

    ### Net input and output flows ###
    
    # Molar flow rates
    @property
    def mol_in(self) -> NDArray[float]:
        """Molar flows going in [kmol/hr]."""
        return sum([s.mol for s in self._ins if s])
    @property
    def mol_out(self) -> NDArray[float]:
        """Molar flows going out [kmol/hr]."""
        return sum([s.mol for s in self._outs if s])

    @property
    def z_mol_in(self) -> NDArray[float]:
        """Molar fractions going in [kmol/hr]."""
        return self.mol_in/self.F_mol_in
    @property
    def z_mol_out(self) -> NDArray[float]:
        """Molar fractions going in."""
        return self.mol_out/self.F_mol_out

    @property
    def F_mol_in(self) -> float:
        """Net molar flow going in [kmol/hr]."""
        return sum([s.F_mol for s in self._ins if s])
    @property
    def F_mol_out(self) -> float:
        """Net molar flow going out [kmol/hr]."""
        return sum([s.F_mol for s in self._outs if s])

    # Mass flow rates
    @property
    def mass_in(self)-> NDArray[float]:
        """Mass flows going in [kg/hr]."""
        return sum([s.mol for s in self._ins if s]) * self._thermo.chemicals.MW
    @property
    def mass_out(self)-> NDArray[float]:
        """Mass flows going out [kg/hr]."""
        return sum([s.mol for s in self._outs if s]) * self._thermo.chemicals.MW

    @property
    def z_mass_in(self)-> NDArray[float]:
        """Mass fractions going in."""
        return self.mass_in/self.F_mass_in
    @property
    def z_mass_out(self)-> NDArray[float]:
        """Mass fractions going out."""
        return self.mass_out/self.F_mass_out

    @property
    def F_mass_in(self)-> float:
        """Net mass flow going in [kg/hr]."""
        return self.mass_in.sum()
    @property
    def F_mass_out(self) -> float:
        """Net mass flow going out [kg/hr]."""
        return self.mass_out.sum()

    # Volumetric flow rates
    @property
    def vol_in(self) -> NDArray[float]:
        """Volumetric flows going in [m3/hr]."""
        return sum([s.vol for s in self._ins if s])
    @property
    def F_vol_in(self) -> float:
        """Net volumetric flow going in [m3/hr]."""
        return sum(self.vol_in)

    @property
    def z_vol_in(self) -> NDArray[float]:
        """Volumetric fractions going in."""
        return self.vol_in/self.F_vol_in
    @property
    def vol_out(self) -> NDArray[float]:
        """Volumetric flows going out [m3/hr]."""
        return sum([s.vol for s in self._outs if s])

    @property
    def F_vol_out(self)-> float:
        """Net volumetric flow going out [m3/hr]."""
        return sum(self.vol_out)
    @property
    def z_vol_out(self) -> NDArray[float]:
        """Volumetric fractions going out."""
        return self.vol_out/self.F_vol_out

    # Enthalpy flow rates
    @property
    def H_in(self) -> float:
        """Enthalpy flow going in [kJ/hr]."""
        return sum([s.H for s in self._ins if s])

    @property
    def H_out(self) -> float:
        """Enthalpy flow going out [kJ/hr]."""
        return sum([s.H for s in self._outs if s])

    @property
    def Hf_in(self) -> float:
        """Enthalpy of formation flow going in [kJ/hr]."""
        return sum([s.Hf for s in self._ins if s])

    @property
    def Hf_out(self) -> float:
        """Enthalpy of formation flow going out [kJ/hr]."""
        return sum([s.Hf for s in self._outs if s])

    @property
    def Hnet(self) -> float:
        """Net enthalpy flow, including enthalpies of formation [kJ/hr]."""
        return self.H_out - self.H_in + self.Hf_out - self.Hf_in


class UnitDesignAndCapital:
    
    __slots__ = (
        'unit',
        'F_BM',
        'F_D',
        'F_P',
        'F_M',
        'design_results',
        'baseline_purchase_costs',
        'purchase_costs',
        'installed_costs',
    )
    
    baseline_purchase_cost = Unit.baseline_purchase_cost
    purchase_cost = Unit.purchase_cost
    installed_cost = Unit.installed_cost
    
    def __init__(self, 
            unit, F_BM, F_D, F_P, F_M,
            design_results: dict,
            baseline_purchase_costs: dict,
            purchase_costs: dict,
            installed_costs: dict,
        ):
        self.unit = unit
        self.F_BM = F_BM
        self.F_D = F_D
        self.F_P = F_P
        self.F_M = F_M
        self.design_results = design_results
        self.baseline_purchase_costs = baseline_purchase_costs
        self.purchase_costs = purchase_costs
        self.installed_costs = installed_costs

    @property
    def equipment_lifetime(self):
        return self.unit.equipment_lifetime



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst
from .auxiliary import Auxiliary
from . import design_tools as design
from .design_tools.specification_factors import (
    pressure_vessel_material_factors,
    material_densities_lb_per_ft3
)
from ..exceptions import bounds_warning, DesignWarning
from warnings import warn

__all__ = ('AuxiliaryPressureVessel',)

class AuxiliaryPressureVessel(Auxiliary):
    __slots__ = ()
    _bounds = design.PressureVessel._bounds
    _units = design.PressureVessel._units
    
    def __init__(self, 
            pressure, diameter, length,
            pressure_units=None, length_units=None, 
            orientation=None, material=None, 
        ):
        super().__init__()
        design_results = self.design_results
        if orientation is None: orientation = 'Vertical'
        if material is None: material = 'Carbon steel'
        design_results['Orientation'] = orientation
        design_results['material'] = material
        if length_units is not None:
            diameter = self.set_design_result('Diameter', length_units, diameter)
            length = self.set_design_result('Length', length_units, length)
        if pressure_units is not None:
            pressure = self.set_design_result('Pressure', pressure_units, pressure)
        rho_M = material_densities_lb_per_ft3[material]
        if pressure < 14.68:
            warn('vacuum pressure vessel ASME codes not implemented yet; '
                 'wall thickness may be inaccurate and stiffening rings may be '
                 'required', category=DesignWarning)
        weight, thickness = design.compute_vessel_weight_and_wall_thickness(
            pressure, diameter, length, rho_M
        )
        self.F_M['Platform and ladders'] = 1.
        design_results['Weight'] = weight  # lb
        design_results['Wall thickness'] = thickness # in
        baseline_purchase_costs = self.baseline_purchase_costs
        if orientation == 'Horizontal':
            key = 'Vertical pressure vessel'
            bounds_warning(self, 'Horizontal vessel weight', weight, 'lb',
                           self._bounds['Horizontal vessel weight'], 'cost')
            bounds_warning(self, 'Horizontal vessel diameter', diameter, 'ft',
                           self._bounds['Horizontal vessel diameter'], 'cost')
            self.F_BM[key] = 3.05
            self.F_M[key] = pressure_vessel_material_factors[material]
            baseline_purchase_costs[key] = design.compute_horizontal_vessel_purchase_cost(weight)
            baseline_purchase_costs['Platform and ladders'] = design.compute_horizontal_vessel_platform_and_ladders_purchase_cost(diameter)
        elif orientation == 'Vertical':
            key = 'Vertical pressure vessel'
            bounds_warning(self, 'Vertical vessel weight', weight, 'lb',
                           self._bounds['Vertical vessel weight'], 'cost')
            bounds_warning(self, 'Vertical vessel length', length, 'ft',
                           self._bounds['Vertical vessel length'], 'cost')
            self.F_BM[key] = 4.16
            self.F_M[key] = pressure_vessel_material_factors[material]
            baseline_purchase_costs[key] = design.compute_vertical_vessel_purchase_cost(weight)
            baseline_purchase_costs['Platform and ladders'] = design.compute_vertical_vessel_platform_and_ladders_purchase_cost(diameter, length)
        else:
            raise ValueError(
                f"invalid orientation {orientation!r}; orientation must be "
                 "either 'Horizontal' or 'Vertical'"
            )
        


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst
from biosteam.units import design_tools as d
from numpy.testing import assert_allclose

def reset_CEPCI(f):
    def g():
        try:
            f()
        finally:
            bst.default_CEPCI()
    g.__name__ = f.__name__
    return g

@reset_CEPCI
def test_electric_motor_cost():
    assert_allclose(
        d.electric_motor_cost(10),
        620.5362044090548, 
    )
    bst.settings.CEPCI = 800
    assert_allclose(
        d.electric_motor_cost(10),
        874.7646934400772, 
    )

@reset_CEPCI
def test_horizontal_vessel_purchase_cost():
    assert_allclose(
        d.compute_horizontal_vessel_purchase_cost(W=1e3),
        8857.95780922862, 
    )
    bst.settings.CEPCI = 800
    assert_allclose(
        d.compute_horizontal_vessel_purchase_cost(W=1e3),
        12486.988982172503,
    )

@reset_CEPCI
def test_vertical_vessel_platform_and_ladders_purchase_cost():
    assert_allclose(
        d.compute_vertical_vessel_platform_and_ladders_purchase_cost(3, 10),
        4708.532188390147, 
    )
    bst.settings.CEPCI = 800
    assert_allclose(
        d.compute_vertical_vessel_platform_and_ladders_purchase_cost(3, 10),
        6637.578415351749,
    )

@reset_CEPCI
def test_vertical_vessel_purchase_cost():
    assert_allclose(
        d.compute_vertical_vessel_purchase_cost(1e3),
        13319.089264026192,
    )
    bst.settings.CEPCI = 800
    assert_allclose(
        d.compute_vertical_vessel_purchase_cost(1e3),
        18775.808654133838,
    )

@reset_CEPCI
def test_purchase_cost_of_trays():
    assert_allclose(
        d.compute_purchase_cost_of_trays(10, 6),
        17092.681096675424,
    )
    bst.settings.CEPCI = 800
    assert_allclose(
        d.compute_purchase_cost_of_trays(10, 6),
        24095.409475489585,
    )

@reset_CEPCI
def test_empty_tower_cost():
    assert_allclose(
        d.compute_empty_tower_cost(3000),
        30829.15740696664,
    )
    bst.settings.CEPCI = 800
    assert_allclose(
        d.compute_empty_tower_cost(3000),
        43459.60515519526,
    )

@reset_CEPCI
def test_plaform_ladder_cost():
    assert_allclose(
        d.compute_plaform_ladder_cost(6, 30),
        16225.103055811393,
    )
    bst.settings.CEPCI = 800
    assert_allclose(
        d.compute_plaform_ladder_cost(6, 30),
        22872.39197294998,
    )

@reset_CEPCI
def test_closed_vessel_turbine_purchase_cost():
    assert_allclose(
        d.compute_closed_vessel_turbine_purchase_cost(power=10),
        15264.970467626266,
    )
    bst.settings.CEPCI = 800
    assert_allclose(
        d.compute_closed_vessel_turbine_purchase_cost(power=10),
        21518.901099737468,
    )
    
if __name__ == '__main__':
    test_electric_motor_cost()
    test_horizontal_vessel_purchase_cost()
    test_vertical_vessel_platform_and_ladders_purchase_cost
    test_vertical_vessel_purchase_cost()
    test_purchase_cost_of_trays()
    test_empty_tower_cost()
    test_plaform_ladder_cost()
    test_closed_vessel_turbine_purchase_cost()


# -*- coding: utf-8 -*-
"""
"""
from dataclasses import dataclass
from thermosteam.utils import AbstractMethod, AbstractClassMethod
from colorpalette import Color
import biosteam as bst

__all__ = ('ProcessModel', 'ScenarioComparison', 'scenario')

def copy(scenario, **kwargs):
    for i in scenario.__slots__:
        if i not in kwargs: kwargs[i] = getattr(scenario, i)
    return scenario.__class__(**kwargs)
    
def scenario_info(scenario, add_metadata):
    slots = scenario.__slots__
    if add_metadata: 
        grey = Color(fg='#878787')
        metadata = scenario.metadata
    arguments = []
    for i in slots:
        j = getattr(scenario, i)
        if isinstance(j, (bool, str)):
            arg = f"{i}={j!r},"
        else:
            try:
                arg = f"{i}={j:.3g},"
            except:
                arg = f"{i}={j},"
        if add_metadata and i in metadata:
            comment = '# ' + metadata[i]
            comment = grey(comment)
            arguments.append(comment)
            arguments.append(arg)
            continue
        arguments.append(
            arg
        )
    if arguments:
        clsname = type(scenario).__name__
        N_spaces = len(clsname) +1
        if N_spaces > 4:
            arguments = '\n    '.join(arguments)
            return (
                f"{clsname}("
                f"\n    {arguments}\n"
                ")"
            
            )
        else:
            spaces = N_spaces * ' '
            arguments = f'\n{spaces}'.join(arguments)
            return (
                f"{clsname}({arguments})"
            
            )
    else:
        return (
            f"{type(scenario).__name__}()"
        )
    
def display_scenario(scenario, metadata=True):
    print(scenario_info(scenario, metadata))
    
def iterate_scenario_data(scenario):
    for i in scenario.__slots__:
        yield (i, getattr(scenario, i))

def scenario_comparison(left, right):
    return ScenarioComparison(left, right)

def scenario(cls):
    if hasattr(cls, 'metadata'): return cls
    cls.metadata = metadata = {}
    for i, j in tuple(cls.__dict__.items()):
        if i.startswith('__'): continue
        if isinstance(j, str):
            if j[0] == '#':
                delattr(cls, i)
                metadata[i] = j.lstrip('# ')
        if isinstance(j, tuple) and len(j) == 2 and isinstance(j[-1], str):
            value, j = j
            if j[0] == '#':
                setattr(cls, i, value)
                metadata[i] = j.lstrip('# ')
            
    # TODO: get defaults and separate units of measure
    cls = dataclass(cls,
        init=True, repr=True, eq=True, unsafe_hash=False, frozen=True,
        match_args=True, slots=True,
    )
    cls.__iter__ = iterate_scenario_data
    cls._ipython_display_ = cls.show = display_scenario
    cls.copy = copy
    cls.__sub__ = scenario_comparison
    return cls

@dataclass(
    init=True, repr=True, eq=True, 
    unsafe_hash=False, frozen=True,
    match_args=True, slots=True,
)
class ScenarioComparison:
    left: object
    right: object

    def show(self):
        parameters = ',\n'.join(['left=' + scenario_info(self.left), 'right=' + scenario_info(self.right)])
        parameters = parameters.replace('\n', '\n    ')
        return print(
            f'{type(self).__name__}(\n'
            f'    {parameters}\n'
            f')'
        )
    _ipython_display_ = show


class ProcessModel:
    """
    ProcessModel objects allow us to write code for many related configurations
    with ease. It streamlines the process of creating a model, including:

    * Defining **scenarios** to compare.
    * Creating the **thermodynamic property package**.
    * Forming the **system** from unit operations.
    * Setting up the evaluation **model**.
    
    Additionally, all objects created within the process model 
    (e.g., chemicals, streams, units, systems) can be easily accessed as attributes.
    
    The first step is to inherit from the ProcessModel abstract class. 
    An abstract class has missing (or "abstract") attributes/methods which 
    the user must define to complete the class. The user must define a 
    `Scenario` dataclass, and `as_scenario`, `create_thermo`, `create_system`,
    `create_model` methods for the process model to initialize its key components. 
    
    It may help to look at how ProcessModel objects are created (approximately):
    
    .. code-block:: python
    
        def __new__(cls, simulate=None, scenario=None, load=True, save=True, **kwargs):
            if scenario is None:
                self.scenario = cls.Scenario(**kwargs)
            else:
                # The Scenario object can be initialized through the `as_scenario` class method.
                self.scenario = cls.as_scenario(scenario)
            
            # No need to recreate a process model for repeated scenarios.
            if load and scenario in cls.cache: return cls.cache[scenario]
            self = super().__new__()
            
            # The thermodynamic property package is given by the `create_thermo` method.
            self.load_thermo(self.create_thermo())
            
            # If no system is returned by the `create_system` method, a new system is created from flowsheet units.
            self.flowsheet = bst.Flowsheet()
            system = self.create_system()
            if system is None: system = self.flowsheet.create_system()
            
            # This saves the system as self.system and all units/streams as attributes by ID.
            # For example, Stream('feedstock') will be stored as self.feestock.
            self.load_system(system) 
            
            # A Model object is loaded from the `create_model` method.
            # The model will be stored as self.model and all parameters and indicators as attributes by function name.
            # For example: 
            #
            # @model.indicator
            # def MSP(): return self.tea.solve_price(self.product)
            #
            # ^ This becomes self.MSP.
            self.load_model(self.create_model())
            
            if simulate: self.system.simulate()
            if save: self.cache[scenario] = self
            return self
    
    """
    #: **class-attribute** Class which defines arguments to the process model using
    #: the layout of a python dataclass: https://docs.python.org/3/library/dataclasses.html
    Scenario: type
    
    #: This method allows the process model to default the scenario.
    #: It should return a Scenario object.
    default_scenario = AbstractMethod
    
    #: This method allows the process model to interpret objects 
    #: (e.g., strings, numbers) as a Scenario.
    as_scenario = AbstractClassMethod
    
    #: This method should return a model object. 
    #: The model will be saved as a self.model attribute. 
    #: All parameters and indicators of the model object will also be saved as 
    #: attributes by their function names.
    initialize = AbstractMethod
    
    #: This method should return a chemicals or thermo object.
    #: BioSTEAM will automatically set it as the thermodynmic property package.
    create_thermo = AbstractMethod
    
    #: This method should create unit operations and connect them.
    #: It can return a system object, optionally. Otherwise, BioSTEAM will 
    #: take care of creating the system from the units and saves 
    #: it as the self.system attribute.
    #: All streams and unit operations are also saved as attributes by their ID.
    create_system = AbstractMethod
    
    #: This method should return a model object. 
    #: The model will be saved as a self.model attribute. 
    #: All parameters and indicators of the model object will also be saved as 
    #: attributes by their function names.
    create_model = AbstractMethod
    
    @classmethod
    def scenario_hook(cls, scenario, kwargs):
        if scenario is None:
            if cls.default_scenario:
                scenario = cls.default_scenario()
            else:
                try:
                    return cls.Scenario(**kwargs)
                except:
                    raise NotImplementedError('missing class method `default_scenario`')
        if not isinstance(scenario, cls.Scenario):
            if cls.as_scenario:
                scenario = cls.as_scenario(scenario)
            else:
                raise NotImplementedError('missing class method `as_scenario`')
        if kwargs: scenario = scenario.copy(**kwargs)
        return scenario
    
    def __init_subclass__(cls):
        cls.cache = {}
        if not hasattr(cls, 'Scenario'):
            cls.Scenario = type('Scenario', (), {})
        if 'Scenario' in cls.__dict__:
            cls.Scenario = scenario(cls.Scenario)
    
    def __new__(cls, *, simulate=True, scenario=None, load=True, save=None, **kwargs):
        scenario = cls.scenario_hook(scenario, kwargs)
        if load and scenario in cls.cache: 
            process_model = cls.cache[scenario]
            if simulate:
                system = process_model.system
                if all([i.isempty() for i in system.products]): system.simulate()
            return process_model
        self = super().__new__(cls)
        self.scenario = scenario
        self.flowsheet = bst.Flowsheet(repr(self))
        thermo = self.create_thermo()
        if thermo is NotImplemented:
            raise NotImplementedError(f"{cls.__name__!r} is missing the `create_thermo` method")
        self.load_thermo(thermo)
        bst.main_flowsheet.set_flowsheet(self.flowsheet)
        unit_registry = self.flowsheet.unit
        unit_registry.open_context_level()
        system = self.create_system()
        if system is NotImplemented:
            raise NotImplementedError(f"{cls.__name__!r} is missing the `create_system` method")
        else:
            units = unit_registry.close_context_level()
            if system is None: system = bst.System.from_units(units=units)
            self.load_system(system)
        model = self.create_model()
        if model is NotImplemented:
            raise NotImplementedError(f"{cls.__name__!r} is missing the `create_model` method")
        elif model is None:
            raise RuntimeError('`create_model` must return a biosteam.Model object')
        self.load_model(model)
        if simulate: system.simulate()
        if save: cls.cache[scenario] = self
        return self
    
    def load_thermo(self, thermo):
        bst.settings.set_thermo(thermo)
        thermo = bst.settings.get_thermo()
        self.chemicals = thermo.chemicals
        self.thermo = thermo
    
    def baseline(self):
        sample = self.model.get_baseline_scenario()
        return sample, self.model(sample)
    
    def load_system(self, system):
        self.system = system
        self.flowsheet = flowsheet = system.flowsheet
        self.__dict__.update(flowsheet.to_dict())
            
    def load_model(self, model):
        self.model = model
        for i in model.parameters:
            setattr(self, i.setter.__name__, i)
            if i.baseline is not None: 
                i.setter(i.baseline)
                i.last_value = i.baseline
        for i in model.optimized_parameters:
            setattr(self, i.setter.__name__, i)
        for i in model.indicators:
            setattr(self, i.getter.__name__, i)
            
    @property
    def parameters(self):
        return self.model._parameters
        
    @property
    def indicators(self):
        return self.model._indicators
    metrics = indicators
    
    def __repr__(self):
        scenario = self.scenario
        scenario_name = type(scenario).__name__
        process_name = type(self).__name__
        N = len(scenario_name)
        return process_name + repr(self.scenario)[N:]
    
    def show(self, metadata=True):
        """Print representation of process model."""
        scenario = self.scenario
        scenario_name = type(scenario).__name__
        process_name = type(self).__name__
        N = len(scenario_name)
        info = scenario_info(scenario, metadata)
        print(process_name + info[N:])
    
    _ipython_display_ = show
    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
__all__ = ('Parameter',)

from ._feature import Feature
from ..utils import format_title
import biosteam as bst
from chaospy import distributions as shape
from inspect import signature

class Parameter(Feature):
    """
    Create a Parameter object that, when called, runs the setter and
    the simulate functions.
    
    Parameters
    ----------
    name : str
        Name of parameter.
    setter : function
        Should set the parameter.
    element : object
        Element associated to parameter.
    system : System
        System associated to parameter.
    distribution : chaospy.Dist
        Parameter distribution.
    units : str
        Units of parameter.
    baseline : float
        Baseline value of parameter.
    bounds : tuple[float, float]
        Lower and upper bounds of parameter.
    coupled : str
        Whether parameter is coupled to the system's mass and energy balances.
        This allows a ConvergenceModel to predict it's impact on recycle loops.
        Defaults to False.
    hook : Callable
        Should return the new parameter value given the sample.
        
    """
    __slots__ = ('setter', 'system', 'distribution', 
                 'baseline', 'bounds', 'coupled', 'hook',
                 'description', 'active', 'last_value')
    
    def __init__(self, name, setter, element, system, distribution,
                 units, baseline, bounds, coupled, hook, description):
        if not name: name, *_ = signature(setter).parameters.keys()
        super().__init__(format_title(name), units, element)
        self.setter = setter.setter if isinstance(setter, Parameter) else setter
        self.system = system
        if isinstance(distribution, str):
            match distribution.lower():
                case 'triangle' | 'triangular':
                    distribution = shape.Triangle(bounds[0], baseline, bounds[1])
                case 'uniform':
                    distribution = shape.Uniform(*bounds)
                case _:
                    raise ValueError(f"invalid distribution {distribution!r}; distribution must be either 'triangular' or 'uniform'")
        elif bounds is None:
            if distribution: bounds = (distribution.lower[0], distribution.upper[0])
        elif not distribution:
            distribution = shape.Uniform(*bounds)
        if bounds and baseline is None:
            baseline = 0.5 * (bounds[0] + bounds[1])
        self.distribution = distribution
        self.baseline = baseline
        self.bounds = bounds
        self.coupled = coupled
        self.hook = hook
        self.description = description
        self.active = True
        self.last_value = None
    
    @classmethod
    def sort_parameters(cls, parameters):
        if not parameters: return
        try:
            system, = set([i.system for i in parameters])
        except:
            raise ValueError('all parameters must have the same system to sort')
        if system is None: return
        unit_path = system.units
        length = len(unit_path)
        def key(parameter):
            if parameter.coupled:
                unit = parameter.unit
                if unit: return unit_path.index(unit) 
            return length
        parameters.sort(key=key) 
    
    @property
    def unit(self):
        """Unit operation directly associated to parameter."""
        element = self.element
        if isinstance(element, bst.Unit): return element
        elif isinstance(element, bst.Stream): return element._sink
    
    @property
    def subsystem(self):
        """Subsystem directly associated to parameter."""
        system = self.system
        if not system:
            return None
        else:
            unit = self.unit
            return system._downstream_system(unit) if unit else system
    
    def simulate(self, **dyn_sim_kwargs):
        """Simulate parameter."""
        if self.coupled:
            subsystem = self.subsystem
            if not subsystem: raise RuntimeError('no system to simulate')
            self.subsystem.simulate(**dyn_sim_kwargs)
        else:
            unit = self.unit
            if hasattr(unit, '_reevaluate'): unit._reevaluate()
    
    def __call__(self, value):
        if self.hook: value = self.hook(value)
        self.setter(value)
        self.last_value = value
        self.simulate()
    
    def _info(self):
        return (f'{type(self).__name__}: {self.name}\n'
             + (f' element: {self.element_name}' if self.element else '')
             + (f' system: {self.system}\n' if self.system else '')
             + (f' units: {self.units}\n' if self.units else '')
             + (f' distribution: {self.distribution}\n' if self.distribution else ''))
    